"""Classes and methods for applying locality-sensitive hashing.

This code is modified from CATCH (https://github.com/broadinstitute/catch/blob/master/catch/utils/lsh.py),
which is released under the following license:
MIT License

Copyright (c) 2018 Hayden Metsky, Broad Institute, Inc., and Massachusetts
Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import defaultdict
import heapq
import logging
import math
import random
import zlib

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class HammingDistanceFamily:
    """An LSH family that works with Hamming distance by sampling bases."""

    def __init__(self, dim):
        self.dim = dim

    def make_h(self):
        """Construct a random hash function for this family.

        Returns:
            hash function
        """
        i = random.randint(0, self.dim - 1)
        def h(x):
            assert len(x) == self.dim
            return x[i]
        return h

    def P1(self, dist):
        """Calculate lower bound on probability of collision for nearby sequences.

        Args:
            dist: Hamming distance; suppose two sequences are within this
                distance of each other

        Returns:
            lower bound on probability that two sequences (e.g., probes) hash
            to the same value if they are within dist of each other
        """
        return 1.0 - float(dist) / float(self.dim)


class HammingWithGUPairsDistanceFamily:
    """An LSH family that works with a pseudo-Hamming distance, which
    tolerates G-U base pairs, by sampling and transforming bases.

    This family can be used with near-neighbor queries if the goal
    is to tolerate G-U pairing between guide and target sequence,
    assuming there are subsequent checks using a distance function
    more conservative (and accurate) than the relaxed one used here.
    However, it should be used with caution with clustering, as the
    distance function used here is relaxed (see below for explanation).

    This family hashes sequences such that similar sequences hash
    to the same value, while tolerating G-U differences. This is
    meant to accommodate G-U base pairing:
    An RNA guide with U can bind to target RNA with G, and vice-versa.
    Note that a guide sequence that ends up being synthesized to RNA
    is the reverse-complement of the guide sequence constructed here, so
    that it will hybridize to the target (here, the guide sequences
    are designed to match the target). If the RNA guide were to have U and
    the RNA target were to have G, then the guide sequence here would be
    A and the target would be G. If the RNA guide were to have G
    and the RNA target were to have U, then the guide sequence here were
    would be C and the target would be T. Thus, to allow G-U pairing,
    we count a base X in the guide sequence as matching a base Y in
    the target if either of the following is true:
      - X == 'A' and Y == 'G'
      - X == 'C' and Y == 'T'

    This is tricky here because the distance function should have
    symmetry -- i.e., the distance between x and y should be equal
    to the distance between y and x. But, given how we defined the
    distance above (in terms of how to count which bases match), this
    would be violated: the distance would depend on whether x is
    a guide sequence or a target sequence (and whether y is a guide
    sequence or target). To solve this, the family of hash functions
    defined here is meant to work with the following distance metric:
      The distance between sequence x and sequence y is the number
      of mismatched bases between x and y after transforming all
      'A' bases in each sequence to 'G' and all 'C' bases in each
      sequence to 'T'. This is equivalent to Hamming distance after
      the 'A'->'G' and 'C'->'T' transformations.

    For example, we will treat the sequences 'AA' and 'GG' as identical
    and ensure they hash to the same value. Likewise, 'CC' and 'TT'
    must hash to the same value. That is, the distance between them
    should be treated as 0 (regardless of which represents a guide
    sequence and which a target).

    This is a more relaxed distance function than what we might need.
    That is, it may report two sequences x and y as being closer to
    each other than we ought to be considering them to be. For example,
    if x represents a guide sequence and x='GG' and y represents a
    target sequence and y='AA', then the distance between x and y
    will be 0 even though we would not want to think of x as binding
    to y. However, this is OK for the purposes of LSH for near-neighbor
    lookups between guide and target sequence, assuming the queries
    have subsequent checks using a more conservative and accurate
    (albeit asymmetric) distance function; the hash family will
    simply lead to more false positive near neighbor results than
    otherwise, but these will still subsequently be pruned by the
    query function.

    The value of P1 is identical to the value defined above for
    (regular) Hamming distance. To see this, let two sequences x
    and y be within distance d of each other, where d is an 'accurate'
    asymmetric distance that accounts for whether x is a guide
    sequence or target (and likewise for y). Let d' be a corresponding
    upper bound on the distance between x and y according to the
    distance function defined above (i.e., after 'A'->'G' and
    'C'->'T' transformations of x and y). Note that d' <= d. The
    probability that x and y hash to different values is <= d'/dim.
    So the probability of collision for x and y, P1(d'), is
    P1(d') >= 1.0 - d'/dim. Since d' <= d, P1(d') >= P1(d). Hence,
    even if d is used as a substitute for d' (as d is the value
    by which we ultimately want to compare x and y), then P1(d)
    is still a valid lower bound on the probability of collision.
    """

    def __init__(self, dim):
        self.dim = dim

    def make_h(self):
        """Construct a random hash function for this family.

        Returns:
            hash function
        """
        i = random.randint(0, self.dim - 1)
        def h(x):
            assert len(x) == self.dim
            b = x[i]
            # Perform 'A'->'G'and 'C'->'T' transformations, as
            # described above; ignore ambiguity codes (for now) and
            # treat them as mismatches
            if b == 'A':
                b = 'G'
            if b == 'C':
                b = 'T'
            return b
        return h

    def P1(self, dist):
        """Calculate lower bound on probability of collision for nearby sequences.

        Args:
            dist: suppose two sequences are within this distance of each other.
                The distance can be either a Hamming distance that does not
                account for G-U pairing, or a Hamming distance after 'A'->'G'
                and 'C'->'T' transformations (see above for explanation of
                why the value returned will be a valid lower bound regardless)

        Returns:
            lower bound on probability that two sequences (e.g., guides) hash
            to the same value if they are within dist of each other
        """
        return 1.0 - float(dist) / float(self.dim)


class MinHashFamily:
    """An LSH family that works by taking the minimum N elements of a
    permutation of k-mers in a string/sequence (MinHash).

    See (Broder et al. 1997) and (Andoni and Indyk 2008) for details.

    The signature produced here also has similarity to Mash (Ondov et al. 2016).
    """

    def __init__(self, kmer_size, N=1):
        """
        Args:
            kmer_size: length of each k-mer to hash
            N: represent the signature of a sequence using hash values of
                the N k-mers in the sequence that have the smallest hash
                values
        """
        self.kmer_size = kmer_size
        self.N = N

    def make_h(self):
        """Construct a random hash function for this family.

        Here, we treat a sequence as being a set of k-mers. We calculate
        a hash value for each k-mer and the hash function on the sequence
        returns the N smallest of these in sorted order.

        Returns:
            hash function
        """
        # First construct a random hash function for a k-mer that
        # is a universal hash function (effectively a "permutation"
        # of the k-mer)
        # Constrain all values to be in [0, 2^31 - 1] to have a bound
        # on the output of the universal hash function; this upper bound
        # is nice because it is also a prime, so we can simply work
        # modulo (2^31 - 1)
        p = 2**31 - 1
        # Let the random hash function be:
        #   (a*x + b) mod p
        # for random integers a, b (a in [1, p] and b in [0, p])
        a = random.randint(1, p)
        b = random.randint(0, p)
        def kmer_hash(x):
            # Hash a k-mer x with the random hash function
            # hash(..) uses a random seed in Python 3.3+, so its output
            # varies across Python processes; use zlib.adler32(..) for a
            # deterministic hash value of the k-mer
            x_hash = zlib.adler32(x.encode('utf-8'))
            return (a * x_hash + b) % p

        def h(s):
            # For a string/sequence s, have the MinHash function be the minimum
            # N hashes, in sorted order, over all the k-mers in it
            assert self.kmer_size <= len(s)
            if self.kmer_size >= len(s) / 2:
                logger.warning(("The k-mer size %d is large (> (1/2)x) "
                    "compared to the size of a sequence to hash (%d), which "
                    "might make it difficult for MinHash to find similar "
                    "sequence"), self.kmer_size, len(s))
            num_kmers = len(s) - self.kmer_size + 1
            if num_kmers < self.N:
                raise ValueError(("The number of k-mers (%d) in the given "
                    "sequence is too small to produce a signature of "
                    "size %d") % (num_kmers, self.N))
            def kmer_hashes():
                for i in range(num_kmers):
                    kmer = s[i:(i + self.kmer_size)]
                    yield kmer_hash(kmer)
            return tuple(sorted(heapq.nsmallest(self.N, kmer_hashes())))
        return h

    def P1(self, dist):
        """Calculate lower bound on probability of collision for nearby sequences.

        Args:
            dist: Jaccard distance (1 minus Jaccard similarity); suppose
                two sequences are within this distance of each other. The
                Jaccard similarity can be thought of as the overlap in k-mers
                between the two sequences

        Returns:
            lower bound on probability that two sequences (e.g., probes) hash
            to the same value if they are within dist of each other
        """
        # With MinHash, the collision probability is the Jaccard similarity
        return 1.0 - dist

    def estimate_jaccard_dist(self, hA, hB):
        """Estimate Jaccard distance between two MinHash signatures.

        Args:
            hA: signature output by h(A) for a hash function h given by
                make_h() and a sequence A (specifically, an ordered list of
                hash values of k-mers in A)
            hB: signature output by h(B) for the same hash function used
                to generate hA, and a sequence B (specifically, an ordered
                list of hash values of k-mers in B)

        Returns:
            estimate of Jaccard distance between signatures A and B
        """
        # Let X = h(hA \union hB). This is equivalent to h(A \union B)
        # where A and B, here, represent the set of k-mers in their
        # respective sequences. X is a random sample of the hashes of
        # the k-mers in (A \union B), and (Y = X \intersect hA \intersect hB)
        # are the hashes in both X and (A \intersect B). So |Y|/|X| is
        # an estimate of the Jaccard similarity. Since X has self.N
        # elements (it is the output of h(.)), this is |Y|/self.N

        # Iterate over hA and hB simultaneously, to count the number of
        # hash values in common between them, among the first self.N of
        # them
        # This is possible because hA and hB are in sorted order
        hA_i, hB_i = 0, 0
        intersect_count = 0
        union_count = 0
        while hA_i < len(hA) and hB_i < len(hB):
            if union_count == self.N:
                # We found all the hash values in X, as defined above
                break
            elif hA[hA_i] < hB[hB_i]:
                hA_i += 1
                union_count += 1
            elif hA[hA_i] > hB[hB_i]:
                hB_i += 1
                union_count += 1
            else:
                # We found a new unique hash value in common:
                # hA[hA_i] == hB[hB_i]
                intersect_count += 1
                union_count += 1
                hA_i += 1
                hB_i += 1

        # It should be that union_count == self.N, except perhaps in
        # some edge cases (e.g., if a hash value is repeated in a signature)
        # And |Y| == intersect_count, where Y is defined above
        similarity = float(intersect_count) / union_count
        return 1.0 - similarity


class HashConcatenation:
    """Concatenated hash functions (AND constructions)."""

    def __init__(self, family, k, join_as_str=False):
        """
        Args:
            family: hash family object; must have a make_h() function
            k: number of hash functions to concatenate
            join_as_str: if True, concatenate the output of the k hash
                functions into a string before returning the concatenated
                result; if False (default), simply return a tuple of the
                k outputs
        """
        self.family = family
        self.k = k
        self.join_as_str = join_as_str
        self.hs = [family.make_h() for _ in range(k)]

    def g(self, x):
        """Evaluate random hash functions and concatenate the result.

        Args:
            x: point (e.g., probe) on which to evaluate hash functions

        Returns:
            concatenation of the result of the self.k random hash functions
            evaluated at x
        """
        if self.join_as_str:
            return ''.join(h(x) for h in self.hs)
        else:
            return tuple([h(x) for h in self.hs])


class NearNeighborLookup:
    """Support for approximate near neighbor lookups.

    This implements the R-near neighbor reporting problem described in
    Andoni and Indyk 2008.
    """

    def __init__(self, family, k, dist_thres, dist_fn, reporting_prob,
                 hash_idx=None, join_concat_as_str=False):
        """
        This selects a number of hash tables (defined as L in the above
        reference) according to the strategy it outlines: we want any
        neighbor (within dist_thres) of a query to be reported with
        probability at least reporting_prob; thus, the number of
        tables should be [log_{1 - (P1)^k} (1 - reporting_prob)]. In
        the above reference, delta is 1.0 - reporting_prob.

        Args:
            family: object giving family of hash functions
            k: number of hash functions from family to concatenate
            dist_thres: consider any two objects within this threshold
                of each other to be neighbors
            dist_fn: function f(a, b) that calculates the distance between
                a and b, to compare against dist_thres
            reporting_prob: report any neighbor of a query with
                probability at least equal to this
            hash_idx: if set, the inserted points are tuples and should
                be key'd on the hash_idx'd index; e.g., (A, B, C) might
                be a point and if hash_idx is 0, it is hashed only based on A,
                B and C simply store additional information along with A,
                and queries are based on distance to A
            join_concat_as_str: if True, have concatenated hash functions
                return a string rather than tuple (this can be more
                efficient, but works only if each hash function from the
                family returns a string)
        """
        self.family = family
        self.k = k
        self.dist_thres = dist_thres
        self.dist_fn = dist_fn
        self.hash_idx = hash_idx

        P1 = self.family.P1(dist_thres)
        if P1 == 1.0:
            # dist_thres might be 0, and any number of hash tables can
            # satisfy the reporting probability
            self.num_tables = 1
        else:
            self.num_tables = math.log(1.0 - reporting_prob, 1.0 - math.pow(P1, k))
            self.num_tables = int(math.ceil(self.num_tables))

        # Setup self.num_tables hash tables, each with a corresponding
        # function for hashing into it (the functions are concatenations
        # of k hash functions from the given family)
        self.hashtables = []
        self.hashtables_masked = []
        self.hashtables_g = []
        for j in range(self.num_tables):
            g = HashConcatenation(self.family, self.k,
                join_as_str=join_concat_as_str)
            self.hashtables += [defaultdict(set)]
            self.hashtables_masked += [defaultdict(set)]
            self.hashtables_g += [g]

    def add(self, pts):
        """Insert given points into each of the hash tables.

        Args:
            pts: collection of points (e.g., probes) to add to the hash
                tables
        """
        for j in range(self.num_tables):
            ht = self.hashtables[j]
            g = self.hashtables_g[j].g
            for p in pts:
                p_key = p[self.hash_idx] if self.hash_idx is not None else p
                ht[g(p_key)].add(p)

    def mask(self, mask_idx, mask_val):
        """Mask points from the hash tables that meet given criteria.

        The points stored must be tuples. This moves the points that meet
        the given criteria (according to mask_idx and mask_val) into separate
        hash tables, so that they can be returned to the main hash tables
        when they should be unmasked.

        Args:
            mask_idx: mask according to this index in a tuple specifying
                a point
            mask_val: mask all points p where p[mask_idx] == mask_val
        """
        for j in range(self.num_tables):
            ht = self.hashtables[j]
            ht_masked = self.hashtables_masked[j]
            keys_to_del = set()
            for k in ht.keys():
                # Find all points p in bucket with key k to mask
                p_to_mask = set()
                for p in ht[k]:
                    if p[mask_idx] == mask_val:
                        p_to_mask.add(p)

                for p in p_to_mask:
                    # Delete p from bucket with key k
                    ht[k].remove(p)
                    # Add p to the mask hash table
                    ht_masked[k].add(p)

                if len(ht[k]) == 0:
                    keys_to_del.add(k)
            # Delete empty buckets in ht
            for k in keys_to_del:
                del ht[k]

    def unmask_all(self):
        """Unmask all points that have been masked.

        This moves all points from the mask hash tables into the main
        hash tables.
        """
        for j in range(self.num_tables):
            ht = self.hashtables[j]
            ht_masked = self.hashtables_masked[j]
            for k in ht_masked.keys():
                for p in ht_masked[k]:
                    ht[k].add(p)

            # Reset this mask hash table
            self.hashtables_masked[j] = defaultdict(set)

    def query(self, q):
        """Find neighbors of a query point.

        Args:
            q: query point (e.g., probe); if self.hash_idx is set and
                the inserted points are tuples, q should only be the
                key of what to search for (i.e., distance is measured
                between q and p[self.hash_idx] for a stored point p)

        Returns:
            collection of stored points that are within self.dist_thres of
            q; all returned points are within this distance, but the
            returned points might not include all that are
        """
        neighbors = set()
        for j in range(self.num_tables):
            ht = self.hashtables[j]
            g = self.hashtables_g[j].g
            for p in ht[g(q)]:
                p_key = p[self.hash_idx] if self.hash_idx is not None else p
                if self.dist_fn(q, p_key) <= self.dist_thres:
                    neighbors.add(p)
        return neighbors




"""Tests for lsh module.

This code is modified from CATCH (https://github.com/broadinstitute/catch/blob/master/catch/utils/tests/test_lsh.py),
which is released under the following license:
MIT License

Copyright (c) 2018 Hayden Metsky, Broad Institute, Inc., and Massachusetts
Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random
import unittest

from adapt.utils import guide
from adapt.utils import lsh

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestHammingDistanceFamily(unittest.TestCase):
    """Tests family of hash functions for Hamming distance.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.HammingDistanceFamily(20)

    def test_identical(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)

        # Identical strings should hash to the same value
        for i in range(30):
            h = self.family.make_h()
            self.assertEqual(h(a), h(b))

    def test_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'ATCGACATGGGCACTGGTAT'

        # a and b should probably collide
        collision_count = 0
        for i in range(10):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertGreater(collision_count, 8)

    def test_not_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'AGTTGTCACCCTTGACGATA'

        # a and b should rarely collide
        collision_count = 0
        for i in range(10):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertLess(collision_count, 2)

    def test_collision_prob(self):
        # Collision probability for 2 mismatches should be
        # 1 - 2/20
        self.assertEqual(self.family.P1(2), 0.9)


class TestHammingWithGUPairsDistanceFamily(unittest.TestCase):
    """Tests family of hash functions for Hamming distance after
    transformations that account for G-U pairing.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.HammingWithGUPairsDistanceFamily(20)

    def test_identical(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)

        # Identical strings should hash to the same value
        for i in range(30):
            h = self.family.make_h()
            self.assertEqual(h(a), h(b))

    def test_identical_gu(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'GCCAATGTGGACGTTACTGC'

        # Identical strings, after 'A'->'G' and 'C'->'T' transformations,
        # should hash to the same value
        for i in range(30):
            h = self.family.make_h()
            self.assertEqual(h(a), h(b))

    def test_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'GTCGGTATGAGCATCGGTGG'

        # a and b should probably collide
        collision_count = 0
        for i in range(10):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertGreater(collision_count, 8)

    def test_not_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'AGTTGTCACCCTTGACGATA'

        # a and b should rarely collide
        collision_count = 0
        for i in range(10):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertLess(collision_count, 2)

    def test_collision_prob(self):
        # Collision probability for 2 mismatches should be
        # 1 - 2/20
        self.assertEqual(self.family.P1(2), 0.9)


class TestMinHashFamilyWithSingleHash(unittest.TestCase):
    """Tests family of hash functions for MinHash, where the hash function
    only returns a single (minimum) value.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.MinHashFamily(3, N=1)

    def test_identical(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)

        # Identical strings should hash to the same value
        h1 = self.family.make_h()
        self.assertEqual(h1(a), h1(b))
        h2 = self.family.make_h()
        self.assertEqual(h2(a), h2(b))

    def test_similar(self):
        a = 'ATCGATATGGGCACTGCTATGTAGCGC'
        b = 'ATCGACATGGGCACTGGTATGTAGCGC'

        # a and b should probably collide; the Jaccard similarity
        # of a and b is ~67% (with 3-mers being the elements that
        # make up each sequence) so they should collide with that
        # probability (check that it is >60%)
        collision_count = 0
        for i in range(100):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertGreater(collision_count, 60)

    def test_not_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'AGTTGTCACCCTTGACGATA'

        # a and b should rarely collide
        collision_count = 0
        for i in range(100):
            h = self.family.make_h()
            if h(a) == h(b):
                collision_count += 1
        self.assertLess(collision_count, 30)

    def test_collision_prob(self):
        # Collision probability for two sequences with a Jaccard
        # distance of 0.2 should be 0.8
        self.assertEqual(self.family.P1(0.2), 0.8)


class TestMinHashFamilySignatures(unittest.TestCase):
    """Tests family of hash functions for MinHash, where the hash function
    returns a signature (multiple hash values).
    """
    
    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.MinHashFamily(4, N=10)

    def test_identical(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)

        # Identical strings should yield the same signature, for the same
        # hash function
        for i in range(10):
            h = self.family.make_h()
            self.assertEqual(h(a), h(b))
            self.assertEqual(self.family.estimate_jaccard_dist(h(a), h(b)), 0.0)

    def test_jaccard_dist_similar(self):
        a = 'ATCGATATGGGCACTGCTATGTAGCGCAAATACGATCGCTAATGCGGATCGGATCGAATG'
        b = 'ATCGACATGGGCACTGGTATGTAGCGCAAATACGATCGCTATTGCGGATCGGATCGAATG'

        # These strings are very similar, but since N is small
        # the Jaccard distance estimate may in some cases be a
        # significant overestimate; test that most of
        # the time, the distance is <=0.5
        num_close = 0
        for i in range(100):
            h = self.family.make_h()
            if self.family.estimate_jaccard_dist(h(a), h(b)) <= 0.5:
                num_close += 1
        self.assertGreaterEqual(num_close, 80)

    def test_jaccard_dist_not_similar(self):
        a = 'ATCGATATGGGCACTGCTATGTAGCGCAAATACGATCGCTAATGCGGATCGGATCGAATG'
        b = 'TCGATCGAATCGAAGGTCGATCGGCGCAATACGGATCGCATTCGATCGGTTATAACGTGA'

        # These strings are far apart, and the estimated Jaccard distance
        # should usually be high
        num_far = 0
        for i in range(100):
            h = self.family.make_h()
            if self.family.estimate_jaccard_dist(h(a), h(b)) > 0.5:
                num_far += 1
        self.assertGreaterEqual(num_far, 80)


class TestHammingHashConcatenation(unittest.TestCase):
    """Tests concatenations of hash functions with Hamming distance.
    """

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.HammingDistanceFamily(20)
        self.G = lsh.HashConcatenation(self.family, 100)
        self.G_join = lsh.HashConcatenation(self.family, 100, join_as_str=True)

    def test_identical(self):
        # Identical a and b should collide even with large k
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)
        self.assertEqual(self.G.g(a), self.G.g(b))
        self.assertEqual(self.G_join.g(a), self.G_join.g(b))

    def test_similar(self):
        # Similar (but not identical) a and b should rarely
        # collide when k is large
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'ATCGACATGGGCACTGGTAT'

        collision_count = 0
        collision_count_join = 0
        for i in range(10):
            if self.G.g(a) == self.G.g(b):
                collision_count += 1
            if self.G_join.g(a) == self.G_join.g(b):
                collision_count_join += 1
        self.assertLess(collision_count, 2)
        self.assertLess(collision_count_join, 2)

    def test_not_similar(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = 'AGTTGTCACCCTTGACGATA'

        # a and b should rarely collide
        collision_count = 0
        collision_count_join = 0
        for i in range(10):
            if self.G.g(a) == self.G.g(b):
                collision_count += 1
            if self.G_join.g(a) == self.G_join.g(b):
                collision_count_join += 1
        self.assertLess(collision_count, 2)
        self.assertLess(collision_count_join, 2)


class TestHammingNearNeighborLookup(unittest.TestCase):
    """Tests approximate near neighbor lookups with Hamming distance."""

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.HammingDistanceFamily(20)
        self.dist_thres = 5
        def f(a, b):
            assert len(a) == len(b)
            return sum(1 for i in range(len(a)) if a[i] != b[i])
        self.dist_fn = f

    def test_varied_k(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)  # identical to a
        c = 'ATCGACATGGGCACTGGTAT'  # similar to a
        d = 'AGTTGTCACCCTTGACGATA'  # not similar to a
        e = 'AGTTGTCACCCTTGACGATA'  # similar to d

        for k in [2, 5, 10]:
            nnl = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95)
            nnl.add([a, b, c, d])

            # b and c are within self.dist_thres of a, so only these
            # should be returned (along with a); note that since
            # a==b, {a,b,c}=={a,c}=={b,c} and nnl.query(a) returns
            # a set, which will be {a,c} or {b,c}
            self.assertCountEqual(nnl.query(a), {a, b, c})

            # Although e was not added, a query for it should return d
            self.assertCountEqual(nnl.query(e), {d})

    def test_masking(self):
        a = ('ATCGATATGGGCACTGCTAT', 1)
        b = (str(a[0]), 2)  # identical to a
        c = ('ATCGACATGGGCACTGGTAT', 3)  # similar to a
        d = ('AGTTGTCACCCTTGACGATA', 1)  # not similar to a
        e = ('AGTTGTCACCCTTGACGATA', 4)  # similar to d

        for k in [2, 5, 10]:
            nnl = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95, hash_idx=0)
            nnl.add([a, b, c, d])

            # As with above test (test_varied_key()), the following should
            # hold true:
            self.assertCountEqual(nnl.query(a[0]), {a, b, c})
            self.assertCountEqual(nnl.query(e[0]), {d})

            # Mask where the second index of the tuple (index 1) is 1
            nnl.mask(1, 1)

            # Now a and d should not be reported
            self.assertCountEqual(nnl.query(a[0]), {b, c})
            self.assertCountEqual(nnl.query(e[0]), {})

            # Unmask all
            nnl.unmask_all()

            # Now the same results as above (before masking) should hold
            self.assertCountEqual(nnl.query(a[0]), {a, b, c})
            self.assertCountEqual(nnl.query(e[0]), {d})


class TestHammingWithGUPairsNearNeighborLookup(unittest.TestCase):
    """Tests approximate near neighbor lookups with Hamming distance
    that supports G-U pairing."""

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        self.family = lsh.HammingWithGUPairsDistanceFamily(20)
        self.dist_thres = 5
        self.dist_fn = guide.seq_mismatches_with_gu_pairs

    def test_varied_k_without_gu_pairs(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)  # identical to a
        c = 'ATCGACATGGGCACTGGTAT'  # similar to a
        d = 'AGTTGTCACCCTTGACGATA'  # not similar to a
        e = 'AGTAGTCACCCTTGACGCTA'  # similar to d

        for k in [2, 5, 10]:
            nnl = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95)
            nnl.add([a, b, c, d])

            # b and c are within self.dist_thres of a, so only these
            # should be returned (along with a); note that since
            # a==b, {a,b,c}=={a,c}=={b,c} and nnl.query(a) returns
            # a set, which will be {a,c} or {b,c}
            self.assertCountEqual(nnl.query(a), {a, b, c})

            # Although e was not added, a query for it should return d
            self.assertCountEqual(nnl.query(e), {d})

    def test_varied_k_with_gu_pairs(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)  # identical to a
        c = 'ACCAACACAAACACCACCAC'  # identical to a in {G,T} space
        d = 'AGTTGTCACCCTTGACGATA'  # not similar to a
        e = 'AGTAGTCACCCTTGACGCTA'  # similar to d
        f = 'AACCACCACGCCCACCACGA'  # similar to d in {G,T} space
        g = 'GGGGGGGGGGTTTTTTTTTT'  # not similar to anything else
        h = 'AAAAAAAAAACCCCCCCCCC'  # similar to g in {G,T} space

        for k in [2, 5, 10]:
            nnl1 = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95)
            nnl1.add([a, b, c, d, g])

            # b is within self.dist_thres of a, so it should be in the
            # output
            # Although c is similar to a in {G,T} space, it is *not*
            # within self.dist_thres of a according to self.dist_fn(a, c),
            # which is what a query for a would compute, so c should not
            # be in the output
            # Note that since a==b, {a,b}=={a}=={b} and nnl.query(a) returns a
            # set, which will be {a} or {b}
            self.assertCountEqual(nnl1.query(a), {a, b})

            # c is within self.dist_thres of a according to self.dist_fn(c, a),
            # which is what a query for c would compute
            # Since a==b, it is also within self.dist_thres of b
            self.assertCountEqual(nnl1.query(c), {a, b, c})

            # Although e was not added, a query for it should return d because
            # they are so similar
            # f is similar to d according to self.dist_fn(f, d), so a
            # query for f should return d
            self.assertCountEqual(nnl1.query(e), {d})
            self.assertCountEqual(nnl1.query(f), {d})

            # h is similar to g according to self.dist_fn(h, g), so a
            # query for h should return g
            self.assertCountEqual(nnl1.query(h), {g})

            # Now make a few changes to what is included in the data
            # structure, and re-test
            nnl2 = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95)
            nnl2.add([a, b, c, d, e, f, h])

            # Since e is now included in the data structure, and f is
            # similar to d according to self.dist_fn(f, d) (as well as to
            # e), a query for f should return both; it should also
            # of course return f since f is included now too
            self.assertCountEqual(nnl2.query(f), {d, e, f})

            # d is *not* similar to f according to self.dist_fn(d, f), so
            # a query for d should not return f (but should return e,
            # and of course d, now that e is included in the data structure)
            self.assertCountEqual(nnl2.query(d), {d, e})

            # Although h is similar to g in {G,T} space, it is *not*
            # similar according to self.dist_fn(g, h), so a query for g
            # should not return h (note that, above, in nnl1, where g is
            # included but not h, a query for h should return g)
            self.assertCountEqual(nnl2.query(g), {})


class TestMinHashNearNeighborLookup(unittest.TestCase):
    """Tests approximate near neighbor lookups with MinHash."""

    def setUp(self):
        # Set a random seed so hash functions are always the same
        random.seed(0)

        kmer_size = 3
        self.family = lsh.MinHashFamily(kmer_size, N=1)
        self.dist_thres = 0.5
        def f(a, b):
            a_kmers = [a[i:(i + kmer_size)] for i in range(len(a) - kmer_size + 1)]
            b_kmers = [b[i:(i + kmer_size)] for i in range(len(b) - kmer_size + 1)]
            a_kmers = set(a_kmers)
            b_kmers = set(b_kmers)
            jaccard_sim = float(len(a_kmers & b_kmers)) / len(a_kmers | b_kmers)
            return 1.0 - jaccard_sim
        self.dist_fn = f

    def test_varied_k(self):
        a = 'ATCGATATGGGCACTGCTAT'
        b = str(a)  # identical to a
        c = 'ATCGACATGGGCACTGGTAT'  # similar to a
        d = 'AGTTGTCACCCTTGACGATA'  # not similar to a
        e = 'AGTTGTCACCCTTGACGATA'  # similar to d

        for k in [2, 5, 10]:
            nnl = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95)
            nnl.add([a, b, c, d])

            # b and c are within self.dist_thres of a, so only these
            # should be returned (along with a); note that since
            # a==b, {a,b,c}=={a,c}=={b,c} and nnl.query(a) returns
            # a set, which will be {a,c} or {b,c}
            self.assertCountEqual(nnl.query(a), {a, b, c})

            # Although e was not added, a query for it should return d
            self.assertCountEqual(nnl.query(e), {d})

    def test_hash_idx(self):
        a = ('ATCGATATGGGCACTGCTAT', 'abc', 'def')
        b = tuple(a)  # identical to a
        c = ('ATCGACATGGGCACTGGTAT', 'ghi', 'jkl')  # similar to a
        d = ('AGTTGTCACCCTTGACGATA', 'mno', 'pqr')  # not similar to a
        e = ('AGTTGTCACCCTTGACGATA', 'stu', 'vwx')  # similar to d

        for k in [2, 5, 10]:
            nnl = lsh.NearNeighborLookup(self.family, k, self.dist_thres,
                self.dist_fn, 0.95, hash_idx=0)
            nnl.add([a, b, c, d])

            # b and c are within self.dist_thres of a, so only these
            # should be returned (along with a); note that since
            # a==b, {a,b,c}=={a,c}=={b,c} and nnl.query(a) returns
            # a set, which will be {a,c} or {b,c}
            self.assertCountEqual(nnl.query(a[0]), {a, b, c})

            # Although e was not added, a query for it should return d
            self.assertCountEqual(nnl.query(e[0]), {d})



"""Functions for working with sequence (and related) i/o.
"""

from collections import defaultdict
from collections import OrderedDict
import gzip
import logging
import re
import textwrap

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def process_fasta(f, replace_degenerate=False,
               skip_gaps=False, make_uppercase=True,
               replace_U=True):
    
    degenerate_pattern = re.compile('[YRWSMKBDHV]')
    m = OrderedDict()
    curr_seq_name = ""
    for line in f:
        line = line.rstrip()
        if len(line) == 0:
            # Reset the sequence being read on an empty line
            curr_seq_name = ""
            continue
        if curr_seq_name == "":
            # Must encounter a new sequence
            assert line.startswith('>')
        if line.startswith('>'):
            curr_seq_name = line[1:]
            m[curr_seq_name] = ''
        else:
            # Append the sequence
            if make_uppercase:
                line = line.upper()
            if replace_degenerate:
                line = degenerate_pattern.sub('N', line)
            if skip_gaps:
                line = line.replace('-', '')
            if replace_U:
                line = line.replace('u', 't').replace('U', 'T')
            m[curr_seq_name] += line

    return m


def read_fasta(fn, replace_degenerate=False,
               skip_gaps=False, make_uppercase=True,
               replace_U=True):
    """Read a FASTA file.

    Args:
        fn: path to FASTA file to read
        replace_degenerate: when True, replace the degenerate
            bases ('Y','R','W','S','M','K','B','D','H','V')
            with 'N'
        skip_gaps: when True, do not read dashes ('-'), which
            represent gaps
        make_uppercase: when True, change all bases to be
            uppercase
        replace_U: when True, change all 'U' bases to 'T'

    Returns:
        dict mapping the name of each sequence to the sequence
        itself. The mapping is ordered by the order in which
        the sequence is encountered in the FASTA file; this
        helps in particular with replicating past results,
        where the input order could affect the output.
    """
    logger.debug("Reading fasta file %s", fn)

    if fn.endswith('.gz'):
        with gzip.open(fn, 'rt') as f:
            m = process_fasta(f, replace_degenerate,
               skip_gaps, make_uppercase)
    else:
        with open(fn, 'r') as f:
            m = process_fasta(f, replace_degenerate,
               skip_gaps, make_uppercase)

    return m


def write_fasta(seqs, out_fn, chars_per_line=70):
    """Write sequences to a FASTA file.

    Args:
        seqs: dict (or OrderedDict) mapping the name of each
            sequence to the sequence itself
        out_fn: path to FASTA file to write
        chars_per_line: the number of characters to put on
            each line when writing a sequence
    """
    with open(out_fn, 'w') as f:
        for seq_name, seq in seqs.items():
            f.write('>' + seq_name + '\n')
            seq_wrapped = textwrap.wrap(seq, chars_per_line,
                break_on_hyphens=False)
            for seq_line in seq_wrapped:
                f.write(seq_line + '\n')
            f.write('\n')


def read_required_guides(fn, expected_guide_length, num_alignments):
    """Read list of required guides.

    There must be 3 columns in the file:
        1 - an identifier for an alignment that the guide should be
            covering (0-based, with maximum value < num_alignments)
        2 - the guide sequence
        3 - the position (start) of the guide sequence in the
            corresponding alignment

    Args:
        fn: path to file, with the format given above
        expected_guide_length: the length of each guide sequence in
            the file in nt; only used as a check
        num_alignments: the number of alignments that the required
            guides cover

    Returns:
        list x of length num_alignments such that x[i] corresponds
        to the i'th alignment, as given in column 1. x[i] is a
        dict mapping guide sequences (column 2) to their position
        (column 2)
    """
    required_guides = [{} for _ in range(num_alignments)]
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')
            aln_id = int(ls[0])
            gd = ls[1]
            gd_pos = int(ls[2])

            # Check aln_id
            if aln_id < 0 or aln_id > num_alignments - 1:
                raise Exception(("Alignment id %d in column 1 of required "
                    "guides file is invalid; must be in [0, %d]") %
                    (aln_id, num_alignments - 1))

            # Check guide length
            if len(gd) != expected_guide_length:
                raise Exception(("Guide with sequence '%s' in required guides "
                    "file has length %d, but it should have length %d") %
                    (gd, len(gd), expected_guide_length))

            # Check that the guide sequence only shows once for this
            # alignment (i.e., has not already appeared)
            if gd in required_guides[aln_id]:
                raise Exception(("Guide with sequence '%s' shows >1 time "
                    "for alignment with id %d") % (gd, aln_id))

            required_guides[aln_id][gd] = gd_pos

    return required_guides


def read_blacklisted_ranges(fn, num_alignments):
    """Read list of blacklisted ranges.

    There must be 3 columns in the file:
        1 - an identifier for an alignment that the guide should be
            covering (0-based, with maxmimum value < num_alignments)
        2 - the start position (inclusive) of a range in the
            corresponding alignment
        3 - the end position (exclusive) of a range in the
            corresponding alignment

    Args:
        fn: path to file, with the format given above
        num_alignments: the number of alignments that the ranges might
            correspond to
   
    Returns:
        list x of length num_alignments such that x[i] corresponds
        to the i'th alignment, as given in column 1. x[i] is a set
        of tuples (start, end) corresponding to the values in columns 2 and 3
    """
    blacklisted_ranges = [set() for _ in range(num_alignments)]
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')
            aln_id = int(ls[0])
            start = int(ls[1])
            end = int(ls[2])

            # Check aln_id
            if aln_id < 0 or aln_id > num_alignments - 1:
                raise Exception(("Alignment id %d in column 1 of blacklisted "
                    "ranges file is invalid; must be in [0, %d]") %
                    (aln_id, num_alignments - 1))

            # Check that end > start
            if start < 0 or end <= start:
                raise Exception(("Blacklisted range [%d, %d) is invalid; "
                    "values must be >= 0 and end > start") % (start, end))

            blacklisted_ranges[aln_id].add((start, end))

    return blacklisted_ranges


def read_blacklisted_kmers(fn, min_len_warning=5, max_len_warning=28):
    """Read file of blacklisted k-mers.

    Args:
        fn: path to FASTA file, where each sequence is a k-mer (names
            of sequences are ignored)
        min_len_warning/max_len_warning: log a warning if the k-mer
            length is outside this range

    Returns:
        set of k-mers
    """
    seqs = read_fasta(fn)
    kmers = set()
    for name, kmer in seqs.items():
        if len(kmer) < min_len_warning:
            logger.warning(("Blacklisted k-mer '%s' might be shorter than "
                "desired and may lead to many guides being treated as "
                "unsuitable") % kmer)
        if len(kmer) > max_len_warning:
            logger.warning(("Blacklisted k-mer '%s' might be longer than "
                "desired") % kmer)
        kmers.add(kmer)
    return kmers


def read_taxonomies(fn):
    """Read file of taxonomies from which to prepare alignments.

    The columns must be, in order:
        1) a label for the row (used for output files; must be unique)
        2) a taxonomic (e.g., species) ID from NCBI
        3) a segment label, or 'None' if unsegmented
        4) one or more accessions of reference sequences (comma-separated)
        5) (optional) metadata filters to only design for a subset of a taxa,
            format: 'metadata=value' or 'metadata!=value', 
                    commas to separate values, semicolons to separate filters

        6) (optional) metadata filters to exclude from a taxa and be specific against,
            column 5 must be included to include this column,
            format: 'metadata=value' or 'metadata!=value', 
                    commas to separate values, semicolons to separate filters

    Args:
        fn: path to TSV file, where each row corresponds to a taxonomy

    Returns:
        list of tuples (label, taxonomic_id, segment, reference_accessions, metadata_filters, metadata_filters_against)
    """
    labels = set()
    taxs = []
    with open(fn) as f:
        for line in f:
            none_strings = ['', 'none']
            ls = line.rstrip().split('\t')
            if len(ls) < 4 or len(ls) > 6:
                raise Exception(("Input taxonomy TSV must have between 4 and 6 columns"))

            label = ls[0]
            if label in labels:
                raise Exception(("Taxonomy label '%s' is not unique") % label)
            labels.add(label)

            try:
                tax_id = int(ls[1])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[1])

            segment = ls[2]
            if segment.lower() in none_strings:
                segment = None
            ref_accs = ls[3].split(',')

            meta_filt = None
            if len(ls) > 4 and ls[4].lower() not in none_strings:
                meta_filt = read_metadata_filters(ls[4].split(';'))

            meta_filt_against = None
            if len(ls) > 5 and ls[5].lower() not in none_strings:
                meta_filt_against = read_metadata_filters(ls[5].split(';'))

            taxs += [(label, tax_id, segment, ref_accs, meta_filt, meta_filt_against)]
    return taxs


def read_accessions_for_taxonomies(fn):
    """Read file of accessions for each taxonomy.

    The columns must be, in order:
        1) a taxonomic (e.g., species) ID from NCBI
        2) a segment label, or 'None' if unsegmented
        3) an accession to include in the design for the taxonomic ID

    A taxonomic ID can appear in multiple rows, if there should be multiple
    accessions used for it.

    Args:
        fn: path to TSV file, where each row corresponds to an accession
            to include in the design

    Returns:
        dict {(taxonomic-id, segment): [list of accessions]}
    """
    accs = defaultdict(list)
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')
            if len(ls) != 3:
                raise Exception(("Input accession TSV must have 3 columns"))

            try:
                tax_id = int(ls[0])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[0])

            segment = ls[1]
            if segment.lower() == 'none':
                segment = None
            accession = ls[2]
            accs[(tax_id, segment)].append(accession)
    return accs


def read_sequences_for_taxonomies(fn):
    """Read sequences for different taxonomies

    The columns must be, in order:
        1) a taxonomic (e.g., species) ID from NCBI
        2) a segment label, or 'None' if unsegmented
        3) a path to a FASTA file containing sequences

    Args:
        fn: path to TSV file, where each row corresponds to a taxonomy

    Returns:
        dict {(taxonomic-id, segment): dict of sequences}
    """
    seqs = defaultdict(list)
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')
            if len(ls) != 3:
                raise Exception(("Input fasta TSV must have 3 columns"))

            try:
                tax_id = int(ls[0])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[0])

            segment = ls[1]
            if segment.lower() == 'none':
                segment = None

            fasta_path = ls[2]
            seqs_for_tax = read_fasta(fasta_path)
            seqs[(tax_id, segment)] = seqs_for_tax
    return seqs


def read_taxonomies_to_design_for(fn):
    """Read different taxonomies to design for.

    The columns must be, in order:
        1) a taxonomic (e.g., species) ID from NCBI
        2) a segment label, or 'None' if unsegmented

    Args:
        fn: path to TSV file, where each row corresponds to a taxonomy

    Returns:
        collection of (taxonomic-id, segment)
    """
    taxs = []
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')
            if len(ls) != 2:
                raise Exception(("Input fasta TSV must have 2 columns"))

            try:
                tax_id = int(ls[0])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[0])

            segment = ls[1]
            if segment.lower() == 'none':
                segment = None

            taxs += [(tax_id, segment)]
    return taxs


def read_taxonomy_specificity_ignore(fn):
    """Read taxonomies to ignore when enforcing specificity.

    The columns must be, in order:
      1) a taxonomic ID for taxonomy A
      2) a taxonomic ID such that this taxonomy should be ignored when
         designing for A, in terms of specificity

    Args:
        fn: path to TSV file

    Returns:
        dict {A: {B}} where {B} consists of the taxonomies that should
        be ignored for A
    """
    tax_ignore = defaultdict(set)
    with open(fn) as f:
        for line in f:
            ls = line.rstrip().split('\t')

            if len(ls) != 2:
                raise Exception(("Input TSV must have 2 columns"))

            try:
                tax_id_a = int(ls[0])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[0])
            try:
                tax_id_b = int(ls[1])
            except ValueError:
                raise Exception(("Taxonomy ID '%s' must be an integer") %
                    ls[1])

            tax_ignore[tax_id_a].add(tax_id_b)
    return tax_ignore


def read_metadata_filters(meta_filts):
    """Create dictionaries of metadata filters from a list.
    
    Args:
        meta_filts: list of filters in the format 'key=values' or 
            'key!=values' with a comma separated list of values

    Returns:
        tuple of 2 dicts where each key is any of ['taxid', 'year', 'country']
            and each value is what to include for the first dict and exclude for
            the second dict
    """
    dict_filter_eq = {}
    dict_filter_neq = {}
    for filt in meta_filts:
        if ' ' in filt:
            raise ValueError("Incorrect format for filter '%s'; individual filters "
                "should not include spaces" %filt)
        dict_to_use = dict_filter_eq
        filt_split = filt.split('!=')
        if len(filt_split) == 2:
            dict_to_use = dict_filter_neq
        else:
            filt_split = filt.split('=')
        if len(filt_split) != 2:
            raise ValueError("Incorrect format for filter '%s'; should be "
                "'metadata=value or metadata!=value'" %filt)
        if filt_split[0] not in ['taxid', 'year', 'country']:
            raise ValueError("Incorrect filter key '%s'; should be one of "
                "['taxid', 'year', 'country']" %filt_split[0])
        if filt_split[0] in ['taxid', 'year']:
            dict_to_use[filt_split[0]] = [int(i) for i in filt_split[1].split(",")]
        else:
            dict_to_use[filt_split[0]] = filt_split[1].split(",")
    return (dict_filter_eq, dict_filter_neq)




"""Classes and methods to shard (or split) k-mers across separate tries.

The basic idea is as follows. Without loss of generality, consider a 28-mer S
that we split into 2 parts.  S has two signatures: s_1 and s_2. s_1 is the
first 14-mer of S collapsed down to a 2-letter alphabet (to account for G-U
pairing). s_2 is the second 14-mer of S collapsed down to a 2-letter alphabet
(to account for G-U pairing). We have a separate trie for every possible
signature, and S gets stored twice: once for s_1 and again for s_2. Now
consider a 28-mer query Q, which similarly has signatures q_1 and q_2. Let's
only query for Q up to 3 mismatches. By the pigeonhole principle, either
Q[0:14] or Q[14:28] has <= 1 mismatch with any possible result. So we can query
in 30 tries (corresponding to 30 signatures): q_1, as well as q_1 where at
every position we flip the letter (i.e., introduce a mismatch); and the same
for q_2. On average, each trie should have 1/2^14 of the total number of
28-mers, so the total space of 28-mers we query is 30/2^14 = 0.002.

In general, consider p partitions (above, p=2) of a k-mer, where p is a small
constant. A loose bound on the query runtime in a single trie is O(# k-mers
in the trie), owing to considerable branching due to G-U pairing. Let n be the
total number of k-mers in the data structure; each trie has on average
n/2^{k/p} k-mers. So a loose bound on the query runtime is
O(p*(n/2^{k/p})*(sum_{i=0}^{floor(m/p)} {k/p choose i})), where m is the number
of mismatches we query up to.
"""

from abc import ABCMeta, abstractmethod
from itertools import combinations

from adapt.specificity import kmer_leaf
from adapt.utils import trie

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TrieSpace:
    """A space of tries, where each trie is addressed by a signature.
    """

    def __init__(self):
        self.tries = {}

    def get(self, sig, make=True):
        """Fetch a trie with a given signature.
        Args:
            sig: signature (e.g., string)
            make: if True, make a trie if one does not exist with sig
        Returns:
            Trie object (or, if one does not exist and make is False, None)
        """
        if sig not in self.tries:
            if make:
                # Make an empty trie, and insert it
                t = trie.Trie()
                self.tries[sig] = t
            else:
                return None
        return self.tries[sig]

    def all_tries(self):
        """Iterate over all tries.
        Yields:
            (signature, trie)
        """
        return self.tries.items()

    def mask(self, d):
        """Mask an object from all leaves in all tries.

        Args:
            d: object to remove from all leaves
        """
        for sig, t in self.all_tries():
            t.mask(d)

    def unmask_all(self):
        """Unmask all masked objects.
        """
        for sig, t in self.all_tries():
            t.unmask_all()


def _signatures_with_mismatches(sig, m):
    """Generate signatures within m mismatches given a signature.

    Args:
        sig: signature as string (alphabet is G and T)
        m: number of mismatches such that all signatures have 0,1,...,m
            mismatches relative to kmer

    Iterates:
        signatures as strings
    """
    yield sig
    for mi in range(1, m+1):
        # Introduce mi mismatches to sig
        for pos in combinations(range(len(sig)), mi):
            # Introduce mismatches to sig at positions in pos
            # Note that len(pos) == mi
            sig_mi = list(str(sig))
            for j in pos:
                # Introduce a mismatch at position j
                if sig_mi[j] == 'G':
                    sig_mi[j] = 'T'
                else:
                    sig_mi[j] = 'G'
            sig_mi_str = ''.join(sig_mi)
            yield sig_mi_str


def _full_signature(kmer):
    """Generate a 2-letter alphabet signature of a full k-mer.

    The transformations are A->G and C->T.

    Args:
        kmer: k-mer

    Returns:
        signature as a string
    """
    return kmer.replace('A', 'G').replace('C', 'T')


def _full_signatures_with_mismatches(kmer, m):
    """Generate signatures within m mismatches given a k-mer.

    Args:
        kmer: k-mer
        m: number of mismatches such that all signatures have 0,1,...,m
            mismatches relative to kmer

    Returns:
        list of signatures as strings
    """
    sig = _full_signature(kmer)
    return _signatures_with_mismatches(sig, m)


def _split_signature(kmer, half):
    """Generate a 2-letter alphabet signature of one half of a k-mer.

    The transformations are A->G and C->T.

    Args:
        kmer: k-mer
        half: 0 or 1, denoting the first or second half

    Returns:
        signature as a string
    """
    k = len(kmer)
    if k % 2 == 0:
        # Make each half be exactly half
        k_half = int(k / 2)
    else:
        # Make the first 'half' be 1 letter longer than the second
        k_half = int(k / 2) + 1
    assert half in [0, 1]
    kmer_half = kmer[0:k_half] if half == 0 else kmer[k_half:k]
    return kmer_half.replace('A', 'G').replace('C', 'T')


def _split_signatures_with_mismatches(kmer, half, m):
    """Generate signatures within one mismatch given a k-mer.

    Args:
        kmer: k-mer, where k is divisible by 2
        half: 0 or 1, denoting the first or second half
        m: number of mismatches such that all signatures have 0,1,...,m
            mismatches relative to kmer

    Returns:
        list of signatures as strings
    """
    sig = _split_signature(kmer, half)
    return _signatures_with_mismatches(sig, m)


class TrieSpaceOfKmers(metaclass=ABCMeta):
    """Abstract class for a space of tries storing k-mers, key'd by signatures
    computed from those k-mers.
    """

    @abstractmethod
    def add(self, kmers): raise NotImplementedError

    @abstractmethod
    def query(self, q, m, gu_pairing): raise NotImplementedError

    @abstractmethod
    def mask(self, taxid): raise NotImplementedError

    @abstractmethod
    def unmask_all(self): raise NotImplementedError


class TrieSpaceOfKmersFullSig(TrieSpaceOfKmers):
    """Space of tries storing k-mers, key'd by the full k-mer.
    """

    def __init__(self):
        self.ts = TrieSpace()
        self.kmer_len = None
        self.contains_kmers = False

    def add(self, kmers):
        """Add to space of tries.

        Args:
            kmers: iterator over (k-mer, {(taxonomy identifier, sequence id)});
                each k-mer must not have ambiguity
        """
        for kmer, seqs_with_kmer in kmers:
            if self.kmer_len is None:
                self.kmer_len = len(kmer)
            if len(kmer) != self.kmer_len:
                raise ValueError("Length of k-mer is variable")
            self.contains_kmers = True

            sig = _full_signature(kmer)
            leaf = kmer_leaf.KmerLeaf(seqs_with_kmer)
            self.ts.get(sig).insert([(kmer, leaf)])

    def query(self, q, m=0, gu_pairing=False):
        """Query a k-mer.

        Args:
            q: query string; must not have ambiguity
            m: number of mismatches to tolerate
            gu_pairing: whether to tolerate G-U base pairing

        Returns:
            dict {taxonomy identifier: {sequence identifiers}}
        """
        if self.contains_kmers is False:
            # Nothing was added; this prevents an error below when
            # self.kmer_len is None if nothing is added
            return {}
        if len(q) != self.kmer_len:
            raise ValueError("Query length must be %d" % self.kmer_len)

        all_results = []
        for sig in  _full_signatures_with_mismatches(q, m):
            t = self.ts.get(sig, make=False)
            if t is None:
                # No results
                pass
            else:
                results = t.query(q, mismatches=m, gu_pairing=gu_pairing)
                all_results.extend(results)
        leaf_union = kmer_leaf.KmerLeaf.union(all_results)
        return leaf_union.d

    def mask(self, taxid):
        """Mask taxid from data structure.

        Args:
            taxid: taxonomic identifier to mask
        """
        self.ts.mask(taxid)

    def unmask_all(self):
        """Unmask all taxonomic identifiers.
        """
        self.ts.unmask_all()


class TrieSpaceOfKmersSplitSig(TrieSpaceOfKmers):
    """Space of tries storing k-mers, key'd by each half of a k-mer.
    """

    def __init__(self):
        # TrieSpace containing first 1/2 of k-mer, forward direction
        self.ts_0 = TrieSpace()
        # TrieSpace containing second 1/2 of k-mer, reverse direction
        self.ts_1 = TrieSpace()

        self.kmer_len = None
        self.contains_kmers = False

    def add(self, kmers):
        """Add to space of tries.

        Args:
            kmers: iterator over (k-mer, {(taxonomy identifier, sequence id)})
                where the latter describes the sequences containing the k-mer;
                each k-mer must not have ambiguity
        """
        for kmer, seqs_with_kmer in kmers:
            if self.kmer_len is None:
                self.kmer_len = len(kmer)
            if len(kmer) != self.kmer_len:
                raise ValueError("Length of k-mer is variable")
            self.contains_kmers = True

            sig_0 = _split_signature(kmer, 0)
            sig_1 = _split_signature(kmer, 1)
            kmer_rev = kmer[::-1]

            leaf_0 = kmer_leaf.KmerLeaf(seqs_with_kmer)
            leaf_1 = kmer_leaf.KmerLeaf(seqs_with_kmer)
            self.ts_0.get(sig_0).insert([(kmer, leaf_0)])
            self.ts_1.get(sig_1).insert([(kmer_rev, leaf_1)])
    
    def query(self, q, m=0, gu_pairing=False):
        """Query a k-mer.

        Args:
            q: query string; must not have ambiguity
            m: number of mismatches to tolerate
            gu_pairing: whether to tolerate G-U base pairing

        Returns:
            dict {taxonomy identifier: {sequence identifiers}}
        """
        if self.contains_kmers is False:
            # Nothing was added; this prevents an error below when
            # self.kmer_len is None if nothing is added
            return {}
        if len(q) != self.kmer_len:
            raise ValueError("Query length must be %d" % self.kmer_len)

        # By the pigeonhole principle, there will be at most floor(m/2)
        # mismatches in one of the two halves of the query
        m_half = int(m / 2)

        # Determine the signatures of each half of the query to use

        if m_half == 0:
            sigs_0 = [_split_signature(q, 0)]
            sigs_1 = [_split_signature(q, 1)]
        else:
            sigs_0 = _split_signatures_with_mismatches(q, 0, m_half)
            sigs_1 = _split_signatures_with_mismatches(q, 1, m_half)

        q_rev = q[::-1]

        all_results = []

        # Query for tries maybe containing each half of the query
        for reverse, sigs, ts in [(False, sigs_0, self.ts_0), (True, sigs_1,
                self.ts_1)]:
            q_q = q_rev if reverse else q

            # Only allow up to m_half mismatches for the first
            # half of the search down the trie
            if self.kmer_len % 2 == 0:
                # The first half of the search down the trie contains
                # levels 0 through k/2 - 1 (inclusive), where k is
                # k-mer length
                first_half_level = int(self.kmer_len/2 - 1)
            else:
                # Since the k-mer length is odd, the first 'half' of
                # the k-mer is 1 letter longer than the second 'half'
                # of the k-mer
                if reverse:
                    # The first half of the search is intended to
                    # match the second 'half' of the k-mer; this
                    # is levels 0 through floor(k/2) - 1 (inclusive)
                    first_half_level = int(self.kmer_len/2) - 1
                else:
                    # The first half of the search is intended to
                    # match the first 'half' of the k-mer; this is
                    # levels 0 through floor(k/2) (inclusive)
                    first_half_level = int(self.kmer_len/2)

            # Lookup each signature, and query the resulting trie (if there
            # is one) for each
            for sig in sigs:
                t = ts.get(sig, make=False)
                if t is None:
                    # No results
                    pass
                else:
                    results = t.query(q_q,
                            mismatches=m, gu_pairing=gu_pairing,
                            mismatches_to_level=(m_half, first_half_level))
                    all_results.extend(results)

        leaf_union = kmer_leaf.KmerLeaf.union(all_results)
        return leaf_union.d

    def mask(self, taxid):
        """Mask taxid from data structure.

        Args:
            taxid: taxonomic identifier to mask
        """
        self.ts_0.mask(taxid)
        self.ts_1.mask(taxid)

    def unmask_all(self):
        """Unmask all taxonomic identifiers.
        """
        self.ts_0.unmask_all()
        self.ts_1.unmask_all()




"""Data structure for leaf node of tries.

The leaf nodes should support specificity queries.
"""

from adapt.utils import trie

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class KmerLeaf(trie.LeafInfo):
    """Store info in each leaf node on the k-mer leading to the leaf.

    This stores a dict, d: {taxonomy identifier: {sequence identifiers}}
    giving the taxonomies and sequences containing the k-mer.
    """
    def __init__(self, xs):
        """
        Args:
            xs: collection of tuples (a, b) where a gives taxonomy identifier
                and b gives sequence identifier for a sequence containing
                this k-mer
        """
        self.d = {}
        for x in xs:
            taxid, seqid = x
            if taxid not in self.d:
                self.d[taxid] = set()
            self.d[taxid].add(seqid)

    def extend(self, other):
        """Extend self to contain what is in other.

        Args:
            other: KmerLeaf object
        """
        for taxid, seqids in other.d.items():
            if taxid not in self.d:
                self.d[taxid] = set()
            self.d[taxid].update(seqids)

    def remove(self, taxid):
        """Remove taxonomy identifier from self.

        Args:
            taxid: taxonomy identifier
        """
        if taxid in self.d:
            del self.d[taxid]

    def is_empty(self):
        """Determine if self has any information stored.

        Returns:
            True iff self.d is empty
        """
        return len(self.d) == 0

    def __contains__(self, taxid):
        """Determine if a taxonomy identifier is stored by this leaf.

        Args:
            taxid: taxonomy identifier

        Returns:
            True iff taxid is in self.d
        """
        return taxid in self.d

    def copy(self):
        """Copy self.

        Returns:
            KmerLeaf object, identical to this one
        """
        d_new = {}
        for taxid in self.d.keys():
            d_new[taxid] = set(self.d[taxid])
        new = KmerLeaf({(-1,-1)})
        new.d = d_new
        return new

    def __eq__(self, other):
        """Check equality of self and other.

        Args:
            other: KmerLeaf object

        Returns:
            True or False
        """
        return self.d == other.d

    def __len__(self):
        """Sum number of sequences stored.

        Returns:
            total number of sequences stored across taxids
        """
        n = 0
        for taxid, seqids in self.d.items():
            n += len(seqids)
        return n

    @staticmethod
    def union(kmer_leaves):
        """Construct a KmerLeaf object representing a union of others.

        Args:
            kmer_leaves: collection of KmerLeaf objects

        Returns:
            KmerLeaf object
        """
        union = KmerLeaf([])
        for kl in kmer_leaves:
            union.extend(kl)
        return union




"""Methods for analyzing coverage of a design against a collection of sequences.
"""

from collections import defaultdict
import logging

import numpy as np

from adapt.utils import guide

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class Design:
    """Immutable object representing a design.

    This can represent just a collection of guide sequences, or a combination
    of primers and guides between those.
    """

    def __init__(self, guides, primers=None):
        """
        Args:
            guides: collection of guide sequences
            primers: if set, a tuple (l, r) where l is a collection of primers
                on the 5' end and r is a collection of primers on the 3' end
        """
        self.guides = guides
        self.primers = primers

        self.is_complete_target = primers is not None


class CoverageAnalyzer:
    """Methods to analyze coverage of a design.
    """

    def __init__(self, seqs, designs, primer_mismatches=None,
            fully_sensitive=False):
        """
        Args:
            seqs: dict mapping sequence name to sequence; checks coverage of the
                designs against these sequences
            designs: dict mapping an identifier for a design to Design object
            primer_mismatches: number of mismatches to tolerate when determining
                whether a primer hybridizes to a target sequence (if not
                set, the designs are assumed to not contain primers)
            fully_sensitive: if True, use a fully sensitive lookup of
                primers and guides (slower)
        """
        self.seqs = seqs
        self.designs = designs
        self.primer_mismatches = primer_mismatches
        self.seqs_are_indexed = False
        self.seqs_index_k = None
        self.fully_sensitive = fully_sensitive

    def _index_seqs(self, k=6, stride_by_k=False):
        """Construct index of seqs for faster lookups of binding positions.

        This constructs a dictionary d. For key 'name' in self.seqs,
        d[name][kmer] provides a collection of positions in self.seqs[name]
        that contain kmer.

        Args:
            k: length of k-mer to index
            stride_by_k: if False, index the position of every single k-mer
                in each sequence; if True, only index k-mers while striding
                by k (i.e., index every k'th k-mer). Enabling this can
                save memory but may make it less likely to find a hit
        """
        logger.info(("Indexing sequences"))
        self.seqs_index_k = k
        self.seqs_index = {}
        for name, seq in self.seqs.items():
            self.seqs_index[name] = defaultdict(set)
            i = 0
            while i < len(seq) - k + 1:
                kmer = seq[i:(i+k)]
                self.seqs_index[name][kmer].add(i)
                if stride_by_k:
                    i += k
                else:
                    i += 1
        self.seqs_are_indexed = True
        logger.info(("Done indexing sequences"))

    def evaluate_pos_by_mismatches(self, seq, target_seq, pos,
            mismatches, allow_gu_pairs):
        """Evaluate binding with a mismatch model.

        Args:
            seq: guide or primer sequence to lookup
            target_seq: target sequence against which to check
            pos: list of positions indicating subsequence in target_seq
                to check for binding of seq
            mismatches: number of mismatches to tolerate when determining
                whether seq binds to target_seq at a position
            allow_gu_pairs: if True, tolerate G-U base pairs between
                seq and a subsequence of target_seq

        Returns:
            set of positions in pos where seq binds to target_seq
        """
        bind_pos = set()
        for i in pos:
            target_subseq = target_seq[i:(i + len(seq))]
            if guide.guide_binds(seq, target_subseq, mismatches,
                    allow_gu_pairs):
                bind_pos.add(i)
        return bind_pos

    def guide_bind_fn(self, seq, target_seq, pos):
        """Determine whether guide detects target at positions.

        This should be implemented in a subclass.
        """
        raise NotImplementedError()

    def find_binding_pos(self, target_seq_name, seq, bind_fn,
            save_activities=None):
        """Find if and where a sequence binds to in a target sequence.

        This takes a naive (but fully sensitive) approach: it simply
        slides across the target sequence, checking for binding at
        each position.

        This can also use a precomputed index of the target sequences to
        speed up finding binding positions. If there are hits from this
        index, it may only report these (depending on allow_not_all_positions).
        But if there are no hits from the index, this will also fall
        back on the naive sliding approach.

        Args:
            target_seq_name: name of target sequence against which to check
                (target sequence is self.seqs[target_seq_name])
            seq: guide or primer sequence to lookup
            bind_fn: function accepting (seq, target_seq, pos) and outputs
                positions in pos to which seq binds to target_seq
            save_activities: if set, pass along to guide_fn to save computed
                activities

        Returns:
            collection of start positions in target_seq to which seq binds;
            empty collection if seq does not bind anywhere in target_seq
        """
        if self.fully_sensitive:
            # Use a naive, slow sliding approach
            allow_not_all_positions = False
            allow_not_fully_sensitive = False
        else:
            # Allow the returned collection of positions to not include *all*
            # positions that seq binds to
            allow_not_all_positions = True
            # Allow returning no positions based only on k-mer lookups in the
            # index (i.e., without performing a slow sliding approach)
            allow_not_fully_sensitive = True

        if not self.seqs_are_indexed:
            self._index_seqs()

        def bind_fn_with_save(seq, target_seq, pos):
            if save_activities is None:
                return bind_fn(seq, target_seq, pos)
            else:
                return bind_fn(seq, target_seq, pos,
                        save_activities=save_activities)

        target_seq = self.seqs[target_seq_name]

        bind_pos = set()
        if allow_not_all_positions:
            # Use the index of the target sequences to find potential hits,
            # and only report these, but looking up every k-mer in seq
            k = self.seqs_index_k
            index = self.seqs_index[target_seq_name]
            for j in range(len(seq) - k + 1):
                kmer = seq[j:(j+k)]
                for kmer_start in index[kmer]:
                    if kmer_start - j in bind_pos:
                        # This is already a binding positon; don't
                        # bother checking it again
                        continue
                    # The k-mer starts at kmer_start, so we have to look
                    # backwards to find the potential start of seq within
                    # target_seq
                    if (kmer_start - j >= 0 and
                            kmer_start - j + len(seq) <= len(target_seq)):
                        pos_to_evaluate = [kmer_start - j]
                        bind_pos.update(bind_fn_with_save(seq, target_seq,
                            pos_to_evaluate))
            if len(bind_pos) > 0:
                # We have found at least 1 binding position; this may
                # not be all but we do not need to find all
                return bind_pos
            if allow_not_fully_sensitive:
                # Return bind_pos (an empty set) even without performing
                # the fully sensitive approach
                return bind_pos
            # If len(bind_pos) == 0, then do the fully sensitive naive
            # sliding approach

        pos_to_evaluate = list(range(len(target_seq) - len(seq) + 1))
        bind_pos = bind_fn_with_save(seq, target_seq, pos_to_evaluate)
        return bind_pos

    def seqs_where_guides_bind(self, guide_seqs):
        """Determine sequences to which guide binds.

        Args:
            guide_seqs: guide sequences to lookup

        Returns:
            collection of names of sequences in self.seqs to which a guide
            in guide_seqs binds
        """
        seqs_bound = set()
        for seq_name, target_seq in self.seqs.items():
            for guide_seq in guide_seqs:
                bind_pos = self.find_binding_pos(seq_name, guide_seq,
                        self.guide_bind_fn)
                if len(bind_pos) > 0:
                    # guide_seq binds somewhere in target_seq
                    seqs_bound.add(seq_name)
                    break
        return seqs_bound

    def seqs_where_targets_bind(self, guide_seqs, primer_left_seqs,
           primer_right_seqs):
        """Determine sequences to which a target binds.

        Args:
            guide_seqs: guide sequences in the target
            primer_left_seqs: primers on the 5' end of the target
            primer_right_seqs: primers on the 3' end of the target

        Returns:
            collection of names of sequences in self.seqs to which
            some pair of primers and guide between them binds
        """
        def primer_bind_fn(seq, target_seq, pos):
            return self.evaluate_pos_by_mismatches(seq, target_seq, pos,
                self.primer_mismatches, False)

        seqs_bound = set()
        for seq_name, target_seq in self.seqs.items():
            guide_bind_pos = set()
            min_guide_len_at_pos = {}
            for guide_seq in guide_seqs:
                bind_pos = self.find_binding_pos(seq_name, guide_seq,
                        self.guide_bind_fn)
                guide_bind_pos.update(bind_pos)
                for pos in bind_pos:
                    if pos not in min_guide_len_at_pos:
                        min_guide_len_at_pos[pos] = len(guide_seq)
                    else:
                        min_guide_len_at_pos[pos] = min(
                                min_guide_len_at_pos[pos], len(guide_seq))
            primer_left_bind_pos = set()
            min_primer_len_at_pos = {}
            for primer_seq in primer_left_seqs:
                bind_pos = self.find_binding_pos(seq_name, primer_seq,
                        primer_bind_fn)
                primer_left_bind_pos.update(bind_pos)
                for pos in bind_pos:
                    if pos not in min_primer_len_at_pos:
                        min_primer_len_at_pos[pos] = len(primer_seq)
                    else:
                        min_primer_len_at_pos[pos] = min(
                                min_primer_len_at_pos[pos], len(primer_seq))
            primer_right_bind_pos = set()
            for primer_seq in primer_right_seqs:
                bind_pos = self.find_binding_pos(seq_name, primer_seq,
                        primer_bind_fn)
                primer_right_bind_pos.update(bind_pos)

            # Determine if there exists some combination of (5' primer)/guide/
            # (3' primer) that binds to target_seq
            added_seq = False
            for pl_pos in primer_left_bind_pos:
                # Consider all primers that fall *after* (3' of) pl
                primer_len = min_primer_len_at_pos[pl_pos]
                for pr_pos in primer_right_bind_pos:
                    if pr_pos > pl_pos:
                        # Find if any guide falls in between pl and pr
                        for g_pos in guide_bind_pos:
                            guide_len = min_guide_len_at_pos[g_pos]
                            if (g_pos >= pl_pos + primer_len and
                                    pr_pos >= g_pos + guide_len):
                                seqs_bound.add(seq_name)
                                added_seq = True
                                break
                    if added_seq:
                        break
                if added_seq:
                    break

        return seqs_bound

    def seqs_bound_by_design(self, design):
        """Determine the sequences bound by a Design.

        Args:
            design: Design object

        Returns:
            collection of sequences that are bound by the guides (and
            possibly primers) in design
        """
        if design.is_complete_target:
            primers_left, primers_right = design.primers
            return self.seqs_where_targets_bind(design.guides,
                    primers_left, primers_right)
        else:
            return self.seqs_where_guides_bind(design.guides)

    def frac_of_seqs_bound(self):
        """Determine the fraction of sequences bound by each design.

        Returns:
            dict mapping design identifier (self.designs.keys()) to the
            fraction of all sequences that are bound by the design it
            represents
        """
        frac_bound = {}
        for design_id, design in self.designs.items():
            logger.info(("Computing fraction of sequences bound by design "
                "'%s'"), str(design_id))
            seqs_bound = self.seqs_bound_by_design(design)
            frac_bound[design_id] = float(len(seqs_bound)) / len(self.seqs)
        return frac_bound

    def mean_activity_of_guides(self):
        """Determine the mean activity of guides from each design.

        Returns:
            dict mapping design identifier (self.designs.keys()) to the
            mean activity, across the target sequences, of its guide set
        """
        mean_activities = {}
        for design_id, design in self.designs.items():
            logger.info(("Computing mean activities of guides in design "
                "'%s'"), str(design_id))
            activities_across_targets = self.activities_where_guide_binds(design.guides)
            mean_activities[design_id] = np.mean(list(activities_across_targets.values()))
        return mean_activities


class CoverageAnalyzerWithMismatchModel(CoverageAnalyzer):
    """Methods to analyze coverage of a design using model with fixed number
    of mismatches for guide-target binding.
    """

    def __init__(self, seqs, designs, guide_mismatches, primer_mismatches=None,
            allow_gu_pairs=True, fully_sensitive=False):
        """
        Args:
            seqs: dict mapping sequence name to sequence; checks coverage of the
                designs against these sequences
            designs: dict mapping an identifier for a design to Design object
            guide_mismatches: number of mismatches to tolerate when determining
                whether a guide hybridizes to a target sequence
            primer_mismatches: number of mismatches to tolerate when determining
                whether a primer hybridizes to a target sequence (if not
                set, the designs are assumed to not contain primers)
            allow_gu_pairs: if True, tolerate G-U base pairs between a guide
                and target when computing whether a guide binds
            fully_sensitive: if True, use a fully sensitive lookup of
                primers and guides (slower)
        """
        super().__init__(seqs, designs, primer_mismatches=primer_mismatches,
                fully_sensitive=fully_sensitive)
        self.guide_mismatches = guide_mismatches
        self.allow_gu_pairs = allow_gu_pairs

    def activities_where_guide_binds(self, guide_seqs):
        raise NotImplementedError()

    def guide_bind_fn(self, seq, target_seq, pos):
        """Evaluate binding with a mismatch model.

        Args:
            seq: guide or primer sequence to lookup
            target_seq: target sequence against which to check
            pos: list of positions indicating subsequence in target_seq
                to check for binding of seq

        Returns:
            set of positions in pos where seq binds to target_seq
        """
        return self.evaluate_pos_by_mismatches(seq, target_seq,
                pos, self.guide_mismatches, self.allow_gu_pairs)


class CoverageAnalyzerWithPredictedActivity(CoverageAnalyzer):
    """Methods to analyze coverage of a design using model that determines
    guide-target binding based on whether it is predicted to be active or
    highly active.
    """

    def __init__(self, seqs, designs, predictor, primer_mismatches=None,
            highly_active=False, fully_sensitive=False):
        """
        Args:
            seqs: dict mapping sequence name to sequence; checks coverage of the
                designs against these sequences
            designs: dict mapping an identifier for a design to Design object
            predictor: adapt.utils.predict_activity.Predictor object, used to
                determine whether a guide-target pair is predicted to be
                active
            primer_mismatches: number of mismatches to tolerate when determining
                whether a primer hybridizes to a target sequence (if not
                set, the designs are assumed to not contain primers)
            highly_active: if True, determine a guide-target pair to bind if
                 it is predicted to be highly active (not just active)
            fully_sensitive: if True, use a fully sensitive lookup of
                primers and guides (slower)
        """
        super().__init__(seqs, designs, primer_mismatches=primer_mismatches,
                fully_sensitive=fully_sensitive)
        self.predictor = predictor
        self.highly_active = highly_active

    def activities_where_guide_binds(self, guide_seqs):
        """Determine activities across the target sequences.

        Args:
            guide_seqs: guide sequences to lookup (treat as a guide set)

        Returns:
            dict {target seq in self.seqs: predicted activity}
        """
        target_act = {}
        for seq_name, target_seq in self.seqs.items():
            activities_for_target = []
            for guide_seq in guide_seqs:
                activities = {}
                bind_pos = self.find_binding_pos(seq_name, guide_seq,
                        self.guide_bind_fn, save_activities=activities)
                if len(bind_pos) > 0:
                    # guide_seq binds somewhere in target_seq
                    # As its activity, take the maximum over multiple
                    # positions (if there is more than one where it may bind)
                    activities_for_target += [max(activities.values())]
                else:
                    activities_for_target += [0]
            # If there are multiple guides, take the max across them
            target_act[target_seq] = max(activities_for_target)
        return target_act

    def guide_bind_fn(self, seq, target_seq, pos, save_activities=None):
        """Evaluate binding with a predictor -- i.e., based on what is
        predicted to be highly active.

        Args:
            seq: guide or primer sequence to lookup
            target_seq: target sequence against which to check
            pos: list of positions indicating subsequence in target_seq
                to check for binding of seq
            save_activities: if set, this saves computed activities as
                save_activites[p]=activity for each p in pos for which
                this can compute activity; self.highly_active must be False

        Returns:
            set of positions in pos where seq binds to target_seq
        """
        pairs_to_evaluate = []
        pos_evaluating = []
        for i in pos:
            # Extract a subsequence with context
            extract_start = i - self.predictor.context_nt
            extract_end = i + len(seq) + self.predictor.context_nt
            if (extract_start < 0 or extract_end > len(target_seq)):
                # The context needed for the target to predict activity
                # falls outside the range of target_seq; do not allow binding
                # at this position
                continue
            target_subseq_with_context = target_seq[extract_start:extract_end]
            pairs_to_evaluate += [(target_subseq_with_context, seq)]
            pos_evaluating += [i]

        # Do not bother memoizing predictions because target_seq are not
        # aligned; a position for one target_seq may not correspond to the
        # same position in another
        if self.highly_active:
            predictions = self.predictor.determine_highly_active(-1,
                    pairs_to_evaluate)
            if save_activities is not None:
                raise Exception(("Cannot use save_activities when using "
                    "'highly active' as a criterion"))
        else:
            predictions = self.predictor.compute_activity(-1,
                    pairs_to_evaluate)
            if save_activities is not None:
                for i, p in zip(pos_evaluating, predictions):
                    save_activities[i] = p
            predictions = [bool(p > 0) for p in predictions]
        bind_pos = set()
        for i, p in zip(pos_evaluating, predictions):
            if p is True:
                # subsequence at i is active or highly active
                bind_pos.add(i)
        return bind_pos



"""Tests for seq_io module.
"""

from collections import OrderedDict
import logging
import tempfile
from os import unlink
import unittest

from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestFastaRead(unittest.TestCase):
    """Tests reading a fasta file.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.INFO)

        # Write the temporary fasta file
        self.fasta = tempfile.NamedTemporaryFile(mode='w', delete=False)
        self.fasta.write(">genome_1\n")
        self.fasta.write("ATACG\n")
        self.fasta.write("TATGC\n")
        self.fasta.write(">genome_2\n")
        self.fasta.write("ATCG\n")
        self.fasta.write("TT\n")
        self.fasta.write("Gg\n")
        self.fasta.write("\n")
        self.fasta.write(">genome_3\n")
        self.fasta.write("AAA\n")
        self.fasta.write("CCC\n")
        self.fasta.write("\n")
        self.fasta.write("\n")
        self.fasta.write(">genome_4\n")
        self.fasta.write("AtA\n")
        self.fasta.write("CGC\n")
        self.fasta.write("\n")
        self.fasta.write("\n")
        self.fasta.write("\n")
        self.fasta.write(">genome_5\n")
        self.fasta.write("AGGA\n")
        self.fasta.write("CAAT\n")
        self.fasta.write("\n")
        self.fasta.write("\n")
        # Closes the file so that it can be reopened on Windows
        self.fasta.close()

        self.expected = OrderedDict()
        self.expected["genome_1"] = "ATACGTATGC"
        self.expected["genome_2"] = "ATCGTTGG"
        self.expected["genome_3"] = "AAACCC"
        self.expected["genome_4"] = "ATACGC"
        self.expected["genome_5"] = "AGGACAAT"

    def test_read(self):
        seqs = seq_io.read_fasta(self.fasta.name, make_uppercase=True)
        self.assertEqual(seqs, self.expected)

    def tearDown(self):
        # Delete temporary file
        unlink(self.fasta.name)

        # Re-enable logging
        logging.disable(logging.NOTSET)


class TestFastaWrite(unittest.TestCase):
    """Tests writing and reading a fasta file.
    """

    def setUp(self):
        # Disable logging
        logging.disable(logging.INFO)

        # Create a temporary fasta file
        self.fasta = tempfile.NamedTemporaryFile(mode='w', delete=False)
        # Closes the file so that it can be reopened on Windows
        self.fasta.close()

        self.seqs = OrderedDict()
        self.seqs["genome_1"] = "ATACGTATGC"
        self.seqs["genome_2"] = "ATCGTTGG"
        self.seqs["genome_3"] = "AAACCC"
        self.seqs["genome_4"] = "ATACGC"
        self.seqs["genome_5"] = "AGGACAAT"

    def test_write_and_read(self):
        seq_io.write_fasta(self.seqs, self.fasta.name)
        seqs_read = seq_io.read_fasta(self.fasta.name, make_uppercase=True)
        self.assertEqual(self.seqs, seqs_read)

    def tearDown(self):
        # Delete temporary file
        unlink(self.fasta.name)

        # Re-enable logging
        logging.disable(logging.NOTSET)



"""Classes for querying the specificity of guides against target sequences.
"""

from abc import ABCMeta, abstractmethod
from collections import defaultdict
import logging

from adapt.specificity import kmer_shard
from adapt.utils import guide
from adapt.utils import lsh

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


class AlignmentQuerier(metaclass=ABCMeta):
    """Abstract class supporting queries for potential guide sequences.
    """

    def __init__(self, alns, guide_length, dist_thres, allow_gu_pairs,
                do_not_memoize=False):
        """
        Args:
            alns: list of Alignment objects
            guide_length: length of guide sequences
            dist_thres: detect a queried guide sequence as hitting a
                sequence in an alignment if it is within a distance
                of dist_thres (where the distance is measured with
                guide.seq_mismatches if G-U base pairing is disabled, and
                guide.seq_mismatches_with_gu_pairs if G-U base pairing
                is enabled; effectively these are Hamming distance either
                tolerating or not tolerating G-U base pairing)
            allow_gu_pairs: if True, tolerate G-U base pairs when computing
                whether a guide binds to a target -- i.e., be sensitive to
                G-U pairing when computing mismatches or, in other words,
                do not count a G-U pair as a mismatch
        """
        self.alns = alns
        self.guide_length = guide_length
        self.dist_thres = dist_thres
        self.allow_gu_pairs = allow_gu_pairs
        self.is_setup = False

    @abstractmethod
    def setup(self): raise NotImplementedError

    @abstractmethod
    def mask_aln(self, aln_idx): raise NotImplementedError

    @abstractmethod
    def unmask_all_aln(self): raise NotImplementedError

    @abstractmethod
    def frac_of_aln_hit_by_guide(self, guide): raise NotImplementedError

    def guide_is_specific_to_alns(self, guide, aln_idxs, frac_hit_thres):
        """Determine if guide is specific to a particular alignment.

        Note that this does *not* verify whether guide hits the alignment
        with an index in aln_idx -- only that it does not hit all the others.

        Args:
            guide: guide sequence to check
            aln_idxs: check if guide is specific to all the alignments with
                these indices (self.alns[aln_idxx[i]])
            frac_hit_thres: say that a guide "hits" an alignment A if the
                fraction of sequences in A that it hits is > this value

        Returns:
            True iff guide does not hit alignments other than aln_idx
        """
        frac_of_aln_hit = self.frac_of_aln_hit_by_guide(guide)
        for j, frac in enumerate(frac_of_aln_hit):
            if j in aln_idxs:
                # This frac is for alignment in aln_idxs; it should be high and
                # is irrelevant for specificity (but we won't check that
                # it is high)
                # Note that if aln_idxs[i] has already been masked, then this
                # check should not be necessary (j should never equal
                # aln_idxs[i])
                continue
            if frac > frac_hit_thres:
                # guide hits too many sequences in alignment j
                return False
        return True

    def guide_is_specific_to_alns_fn(self, aln_idxs, frac_hit_thres,
            do_not_memoize=False):
        """Provide a function for determining the specificity of guides.

        This creates and uses a new memoizer.

        Args:
            aln_idxs: check if guide is specific to all the alignments with
                these indices (self.alns[aln_idxx[i]])
            frac_hit_thres: say that a guide "hits" an alignment A if the
                fraction of sequences in A that it hits is > this value
            do_not_memoize: if True, do not memoize query results

        Returns:
            a function with argument guide that returns the output of calling
            self.guide_is_specific_to_alns(guide, ...)
        """
        if do_not_memoize:
            # Just call the function without memoizing
            def gis(guide):
                return self.guide_is_specific_to_alns(
                        guide, aln_idxs, frac_hit_thres)
            return gis

        memoized = {}
        def gis(guide):
            if guide in memoized:
                return memoized[guide]
            else:
                is_spec = self.guide_is_specific_to_alns(
                        guide, aln_idxs, frac_hit_thres)
                memoized[guide] = is_spec
                return is_spec
        return gis


class AlignmentQuerierWithLSHNearNeighbor(AlignmentQuerier):
    """Supports queries for potential guide sequences across a set of
    alignments using near-neighbor lookups with LSH.

    This uses LSH to find near neighbors of queried guide sequences. It
    constructs a data structure containing all subsequences in a given
    collection of alignments (subsequences of length equal to the guide
    length). These are sequences that could potentially be "hit" by a
    guide. Then, on query, it performs approximate near neighbor search
    and calculates the fraction of sequences in each alignment that are
    hit by the queried guide.

    Note that, when building the data structure, this removes all gaps
    from sequences in the alignment. Thus, all subsequences stored are
    gapless. In terms of querying for potential hits, this is more
    realistic because the actual sequences do not have gaps.

    This currently stores in the near-neighbor data structure a tuple
    (subsequence string s, index i of alignment with s) -- i.e., all
    unique subsequences for each alignment. It stores an additional data
    structure that contains, for each of those tuples, all of the
    sequences in alignment i that contain s. Together, these can be used
    to query for all sequences that contain a subsequence that is 'near'
    the queried guide.
    """

    def __init__(self, alns, guide_length, dist_thres, allow_gu_pairs, k=22,
                 reporting_prob=0.95):
        """
        Args:
            [See AlignmentQuerier.__init__()]
            k: number of hash functions to draw from a family of
                hash functions for amplification; each hash function is then
                the concatenation (h_1, h_2, ..., h_k)
            reporting_prob: ensure that any guide within dist_thres of
                a queried guide is detected as such with this probability;
                this constructs multiple hash functions (each of which is a
                concatenation of k functions drawn from the family) to achieve
                this probability
        """
        super().__init__(alns, guide_length, dist_thres, allow_gu_pairs)

        if allow_gu_pairs:
            # Measure distance while tolerating G-U base pairing, and
            # use an appropriate LSH family
            # Note that guide.seq_mismatches_with_gu_pairs is appropriate
            # for the distance function here (although it is asymmetric)
            # because it takes the arguments (guide_seq, target_seq), in
            # that order. This will work with near neighbor queries
            # because lsh.NearNeighborLookup.query(q) calls the dist_fn
            # as dist_fn(q, possible-hit); here, q always represents a
            # potential guide sequence and possible-hit is always a
            # potential target sequence
            dist_fn = guide.seq_mismatches_with_gu_pairs
            family = lsh.HammingWithGUPairsDistanceFamily(guide_length)
        else:
            # Do not tolerate G-U base pairing when measure distance,
            # and use a standard Hamming distance LSH family
            dist_fn = guide.seq_mismatches
            family = lsh.HammingDistanceFamily(guide_length)
        self.nnr = lsh.NearNeighborLookup(family, k, dist_thres, dist_fn,
            reporting_prob, hash_idx=0, join_concat_as_str=True)

        self.seqs_with_subseq = defaultdict(set)

    def setup(self):
        """Build data structure for near neighbor lookup of guide sequences.

        The data structure in self.nnr stores only unique subsequences in
        each alignment; it does not store the particular sequences of
        the alignment that contain the subsequences. A separate hashtable
        stored here, self.seqs_with_subseq, tracks the sequence indices
        in alignment i that contain a subsequence s (key'd on (s, i)).
        """
        for aln_idx, aln in enumerate(self.alns):
            # Convert aln.seqs from column-major order to row-major order
            # Also, remove all gaps from guide sequences because the
            # true sequences are gapless, so queries for potential hits
            # should be against gapless sequence
            seqs = aln.make_list_of_seqs(remove_gaps=True)

            for seq_idx, seq in enumerate(seqs):
                logger.debug(("Indexing for queries: alignment %d of %d, "
                    "sequence %d of %d"),
                    aln_idx + 1, len(self.alns), seq_idx + 1, len(seqs))

                # Add all possible guide sequences g as:
                #   (g, aln_idx)
                pts = set()
                for j in range(len(seq) - self.guide_length + 1):
                    g = seq[j:(j + self.guide_length)]
                    pts.add((g, aln_idx))
                    self.seqs_with_subseq[(g, aln_idx)].add(seq_idx)
                self.nnr.add(pts)
        self.is_setup = True

    def mask_aln(self, aln_idx):
        """Mask an alignment from being reported in near neighbor lookup.

        Note that masking an alignment A can be much more efficient than
        leaving it in the near neighbor data structures and filtering
        the output of queries that match A. This is especially true if
        many neighbors of queries will be from A because the near neighbor
        query function will spend a lot of time comparing the query against
        guides (neighbors) from A.

        Args:
            aln_idx: index of alignment in self.alns to mask from lookups
        """
        logger.debug("Masking alignment with index %d from alignment queries",
            aln_idx)

        # The alignment index is stored in index 1 of the tuple
        mask_idx = 1

        self.nnr.mask(mask_idx, aln_idx)

    def unmask_all_aln(self):
        """Unmask all alignments that may have been masked in the near neighbor
        lookup.
        """
        logger.debug("Unmasking all alignments from alignment queries")

        self.nnr.unmask_all()

    def frac_of_aln_hit_by_guide(self, guide):
        """Calculate how many sequences in each alignment are hit by a query.

        Args:
            guide: guide sequence to query

        Returns:
            list of fractions f where f[i] gives the fraction of sequences
            in the i'th alignment (self.alns[i]) that contain a subsequence
            which is found to be a near neighbor of guide
        """
        if not self.is_setup:
            raise Exception(("AlignmentQuerier.setup() must be called before "
                "querying for near neighbors"))

        assert len(guide) == self.guide_length

        neighbors = self.nnr.query(guide)
        seqs_hit_by_aln = defaultdict(set)
        for neighbor in neighbors:
            _, aln_idx = neighbor
            seq_idxs = self.seqs_with_subseq[neighbor]
            seqs_hit_by_aln[aln_idx].update(seq_idxs)

        frac_of_aln_hit = []
        for i, aln in enumerate(self.alns):
            num_hit = len(seqs_hit_by_aln[i])
            frac_hit = float(num_hit) / aln.num_sequences
            frac_of_aln_hit += [frac_hit]
        return frac_of_aln_hit


class AlignmentQuerierWithKmerSharding(AlignmentQuerier):
    """Supports queries for potential guide sequences across a set of
    alignments using an approach that shards k-mers across many small tries.

    The approach is described in detail in the header of the
    kmer_shard module.

    In contrast to AlignmentQuerierWithLSHNearNeighbor, this is *not*
    probabilistic: it is fully sensitive to finding all hits. There
    are no false positives or false negatives.
    """

    def __init__(self, alns, guide_length, dist_thres, allow_gu_pairs):
        """
        Args:
            [See AlignmentQuerier.__init__()]
        """
        super().__init__(alns, guide_length, dist_thres, allow_gu_pairs)

        # TrieSpaceOfKmersFullSig has smaller tries (owing to the longer
        # signature), so each trie is faster to search. It does require more
        # bit flips -- i.e., possible signatures up to dist_thres mismatches
        # away -- again, owing to the longer signatures. So it will be faster
        # as long as dist_thres is not too high. Once dist_thres is too high,
        # that combinatorial space of mismatches (bit flips in the signatures)
        # dominates, and it is faster to use split signatures -- e.g., with
        # TrieSpaceOfKmersSplitSig. Empirically, TrieSpaceOfKmersFullSig
        # appears faster up to dist_thres <= 4; at dist_thres > 4,
        # TrieSpaceOfKmersSplitSig appears faster.
        if dist_thres <= 4:
            self.tsok = kmer_shard.TrieSpaceOfKmersFullSig()
        else:
            self.tsok = kmer_shard.TrieSpaceOfKmersSplitSig()

    def setup(self):
        """Build data structure for lookup of guide sequences.

        In contrast to AlignmentQuerierWithLSHNearNeighbor, the
        data structure here directly stores the particular
        sequences of the alignment that contain the subsequences; they
        do not have to be stored separately.

        Note that we could call self.tsok.add() each time we see a k-mer with a
        single sequence, but it is faster to add k-mers in batches where each
        k-mer contains most or all of the sequences that contain it (leaf
        information). This produces a batch for each alignment, and calls
        add() once per alignment with all of the sequences containing
        each k-mer in that alignment. Note that inserting into the tries
        will support merging with existing information, so it is ok if
        two alignments contain the same k-mer.

        This leaves out any k-mers with ambiguity because the data structure
        cannot tolerate these. Another option might be to handle them and then
        determine any sequence that *might* hit a k-mer in the data structure
        (up to ambiguity) to be non-specific. However, this alternative option
        does not make sense in the extreme case: for example, if a k-mer is all
        Ns, then every sequence would be deemed non-specific. Nevertheless,
        ignoring k-mers with ambiguity entirely might not be the best choice so
        this should be revisited (TODO).
        """
        unambig_chars = set(['a', 'c', 'g', 't'])
        def is_unambig(kmer):
            return set(kmer.lower()).issubset(unambig_chars)

        for aln_idx, aln in enumerate(self.alns):
            logger.debug(("Indexing for queries: alignment %d of %d"),
                aln_idx + 1, len(self.alns))

            # Convert aln.seqs from column-major order to row-major order
            # Also, remove all gaps from guide sequences because the
            # true sequences are gapless, so queries for potential hits
            # should be against gapless sequence
            seqs = aln.make_list_of_seqs(remove_gaps=True)

            # Find all sequences containing each k-mer
            kmer_seqs = defaultdict(set)
            for seq_idx, seq in enumerate(seqs):
                for j in range(len(seq) - self.guide_length + 1):
                    g = seq[j:(j + self.guide_length)]
                    kmer_seqs[g].add(seq_idx)

            # Produce an iterator over kmer_seqs that adds aln_idx, in
            # the format needed by self.tsok.add(), and add them into
            # the data structure
            def kmers_iter():
                for kmer, seq_idxs in kmer_seqs.items():
                    if not is_unambig(kmer):
                        # Skip k-mers with ambiguity
                        continue
                    yield (kmer, {(aln_idx, seq_idx) for seq_idx in seq_idxs})
            self.tsok.add(kmers_iter())

        self.is_setup = True

    def mask_aln(self, aln_idx):
        """Mask alignment from data structure.

        Args:
            aln_idx: index of alignment in self.alns to mask from lookups
        """
        logger.debug("Masking alignment with index %d from alignment queries",
            aln_idx)

        self.tsok.mask(aln_idx)

    def unmask_all_aln(self):
        """Unmask all alignments that may have been masked.
        """
        logger.debug("Unmasking all alignments from alignment queries")

        self.tsok.unmask_all()

    def frac_of_aln_hit_by_guide(self, guide):
        """Calculate how many sequences in each alignment are hit by a query.

        Args:
            guide: guide sequence to query

        Returns:
            list of fractions f where f[i] gives the fraction of sequences
            in the i'th alignment (self.alns[i]) that contain a subsequence
            which is found to be a hit of guide
        """
        if not self.is_setup:
            raise Exception(("AlignmentQuerier.setup() must be called before "
                "querying for near neighbors"))

        assert len(guide) == self.guide_length

        seqs_hit_by_aln = self.tsok.query(guide, m=self.dist_thres,
                gu_pairing=self.allow_gu_pairs)

        frac_of_aln_hit = []
        for i, aln in enumerate(self.alns):
            num_hit = len(seqs_hit_by_aln[i]) if i in seqs_hit_by_aln else 0
            if aln.num_sequences == 0:
                frac_hit = 0
            else:
                frac_hit = float(num_hit) / aln.num_sequences
            frac_of_aln_hit += [frac_hit]
        return frac_of_aln_hit




"""Functions for working with guides.
"""

import logging

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


# Store the unambiguous bases that make up each
# ambiguous base in the IUPAC notation
FASTA_CODES = {'A': set(('A')),
               'T': set(('T')),
               'C': set(('C')),
               'G': set(('G')),
               'K': set(('G', 'T')),
               'M': set(('A', 'C')),
               'R': set(('A', 'G')),
               'Y': set(('C', 'T')),
               'S': set(('C', 'G')),
               'W': set(('A', 'T')),
               'B': set(('C', 'G', 'T')),
               'V': set(('A', 'C', 'G')),
               'H': set(('A', 'C', 'T')),
               'D': set(('A', 'G', 'T')),
               'N': set(('A', 'T', 'C', 'G'))}

# Specify the substring length to use for checking whether
# to terminate early in guide_binds()
GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN = 5


def seq_mismatches(seq_a, seq_b):
    """Count number of mismatches between two sequences, tolerating ambiguity.

    Note that because 'N' usually signals missing data (and not necessarily
    ambiguity about what the genome actually is), this considers 'N' with
    any other base to be a mismatch.

    Args:
        seq_a: str of a sequence
        seq_b: str of a sequence, same length as seq_a

    Returns:
        number of mismatches between seq_a and seq_b, tolerating ambiguity
        between them (e.g., 'Y' and 'T' is not a mismatch) but considering
        'N' to be a mismatch (e.g., 'N' and 'T' is a mismatch)
    """
    assert len(seq_a) == len(seq_b)

    return sum(1 for i in range(len(seq_a))
               if (seq_a[i] == 'N' or seq_b[i] == 'N' or
                   not (FASTA_CODES[seq_a[i]] & FASTA_CODES[seq_b[i]])))


def seq_mismatches_with_gu_pairs(guide_seq, target_seq):
    """Count number of mismatches between a guide and target, allowing G-U
    pairs, and tolerating ambiguity.

    As in seq_mismatches(..), this considers 'N' with any other base to
    be a mismatch.

    This also tolerates G-U base pairing:
    An RNA guide with U can bind to target RNA with G, and vice-versa.
    Note that a guide sequence that ends up being synthesized to RNA
    is the reverse-complement of the guide sequence constructed here, so
    that it will hybridize to the target (here, the guide sequences
    are designed to match the target). If the RNA guide were to have U and
    the RNA target were to have G, then the guide sequence here would be
    A and the target would be G. If the RNA guide were to have G
    and the RNA target were to have U, then the guide sequence here were
    would be C and the target would be T. Thus, to allow G-U pairing,
    we count a base X in the guide sequence as matching a base Y in
    the target if either of the following is true:
      - X == 'A' and Y == 'G'
      - X == 'C' and Y == 'T'

    Unlike in seq_mismatches(..), it is important to know which sequence
    here represents the guide and which represents the target, so we
    cannot just refer to them interchangeably as seq_a and seq_b.

    Args:
        guide_seq: str of a guide sequence
        target_seq: str of target sequence, same length as guide_seq

    Returns:
        number of mismatches between guide and target sequence, counting
        G-U pairs as matching, and tolerating ambiguity between them
    """
    assert len(guide_seq) == len(target_seq)

    # Since this function is low-level and called so often, it
    # may be better to ensure it is efficient by keeping all
    # conditionals inside the sum(..), even if it is messy, rather
    # than defining and using subfunctions
    return sum(1 for i in range(len(guide_seq)) if (
                   # mismatch if either guide or target is 'N'
                   (guide_seq[i] == 'N' or target_seq[i] == 'N') or
                   # mismatch if not a match
                   not (
                       # both bases match
                       (FASTA_CODES[guide_seq[i]] & FASTA_CODES[target_seq[i]]) or
                       # guide is 'A' and target is 'G'
                       ('A' in FASTA_CODES[guide_seq[i]] and
                        'G' in FASTA_CODES[target_seq[i]]) or
                       # guide is 'C' and target is 'T'
                       ('C' in FASTA_CODES[guide_seq[i]] and
                        'T' in FASTA_CODES[target_seq[i]])
                   )
               )
           )


def query_target_eq(query_seq, target_seq):
    """Determine if a query sequence equals a target.

    This tolerates 'N' in the query sequence but not in the target
    sequence. As a result, it is different than checking if
    seq_mismatches(query_seq, target_seq) == 0. For example,
    a query of 'AN' equals a target of 'AT' (but a query of 'AT'
    does not equal a target of 'AN'). The reason is the same
    as in seq_mismatches() -- 'N' in a target usually signals
    missing data rather than ambiguity.

    Args:
        query_seq: str of a query sequence
        target_seq: str of a target sequence, same length as
            query_seq

    Returns:
        True if query_seq equals target_seq, tolerating all
        non-N ambiguity codes in the query and target, and
        tolerating 'N' in the query; otherwise, False
    """
    assert len(query_seq) == len(target_seq)

    return sum(1 for i in range(len(query_seq))
               if target_seq[i] == 'N' or
               not (FASTA_CODES[query_seq[i]] & FASTA_CODES[target_seq[i]])) == 0
    return num_mismatches == 0


def guide_binds(guide_seq, target_seq, mismatches, allow_gu_pairs):
    """Determine whether a guide binds to a target sequence.

    This tolerates ambiguity and decides whether a guide binds based on
    whether its number of mismatches with the target sequence is within
    a threshold.

    If the target sequence contains a gap (and the guide sequence does
    not, as it should not), this decides that the guide does not bind.

    The calls to seq_mismatches() or seq_mismatches_with_gu_pairs() may
    be slow because they sum over all bases in guide_seq and target_seq;
    in particular, the lookups in FASTA_CODES may take a while for each
    term in the sum. In many cases, the output of those functions (i.e.,
    the number of mismatches between guide_seq and target_seq) will be
    far higher than the given threshold on mismatches. The guide will not
    bind, but we can determine this without having to sum over all bases.
    To speed up these cases (i.e., when guide_seq and target_seq are very
    different), this function calls seq_mismatches() (or
    seq_mismatches_with_gu_pairs()) on just a substring of guide_seq
    and target_seq. The substring is arbitrary: here, it is taken from
    the start of those sequences. If the number of mismatches between
    those substrings is already too high, we can terminate early and
    say that the guide does not bind.

    Args:
        guide_seq: str of a guide sequence
        target_seq: str of a target sequence, same length as guide_seq
        mismatches: int giving threshold on number of mismatches for binding
        allow_gu_pairs: if True, tolerate G-U base pairs when
            counting mismatches between guide_seq and target_seq

    Returns:
        True iff the number of mismatches between guide_seq and target_seq
        is <= mismatches
    """
    if '-' in target_seq:
      assert '-' not in guide_seq
      return False

    if allow_gu_pairs:
        mismatches_fn = seq_mismatches_with_gu_pairs
    else:
        mismatches_fn = seq_mismatches

    # Determine whether to terminate early, using just a substring of
    # the two sequences
    if (mismatches < GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN and
            len(guide_seq) > GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN and
            len(target_seq) > GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN):
        guide_subseq = guide_seq[:GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN]
        target_subseq = target_seq[:GUIDE_BINDS_EARLY_TERMINATE_SUBSTR_LEN]
        m_substr = mismatches_fn(guide_subseq, target_subseq)
        if m_substr > mismatches:
            # Even in just the substring, there are too many mismatches for
            # the guide to bind
            return False

    # Compute the number of mismatches between guide_seq and target_seq,
    # and compare against the given threshold
    m = mismatches_fn(guide_seq, target_seq)
    return m <= mismatches


def gc_frac(guide_seq):
    """Compute fraction of guide that is GC.

    Args:
        guide_seq: string of guide sequence; must be all uppercase

    Returns:
        fraction of guide sequence that is G or C
    """
    gc = guide_seq.count('G') + guide_seq.count('C')
    return float(gc) / len(guide_seq)


def guide_overlap_in_seq(guide_seqs, target_seq, mismatches, allow_gu_pairs):
    """Use a simple sliding strategy to find where guides bind in a sequence.

    This computes binding according to guide_binds().

    This uses a naive sliding strategy over target_seq, computing binding at
    every position. Since this is slow, this assumes target_seq is short (~100s
    of nt). Similarly, this returns a set of indices -- rather than ranges --
    which are easier to work with downstream even though they are less
    efficient.

    Args:
        guide_seqs: list of str of guide sequences
        target_seq: str of a target sequence, at least the length of guide_seq
        mismatches: int giving threshold on number of mismatches for binding
        allow_gu_pairs: if True, tolerate G-U base pairs when
            counting mismatches between guide_seq and target_seq
    
    Returns:
        set of indices in target_seq to which guide_seq binds
    """
    for guide_seq in guide_seqs:
        assert len(target_seq) >= len(guide_seq)

    indices_bound = set()

    for i in range(0, len(target_seq) - len(guide_seq) + 1):
        target_seq_at_i = target_seq[i:(i + len(guide_seq))]
        for guide_seq in guide_seqs:
            binds = guide_binds(guide_seq, target_seq_at_i, mismatches,
                    allow_gu_pairs)
            if binds:
                for j in range(0, len(guide_seq)):
                    indices_bound.add(i + j)

    return indices_bound



"""Tests for design.py
"""

import random
import os
import copy
import tempfile
import unittest
import logging

from collections import OrderedDict
from argparse import Namespace
from adapt import alignment
from adapt.prepare import align, ncbi_neighbors, prepare_alignment
from adapt.utils import seq_io
from bin import design

__author__ = 'Priya Pillai <ppillai@broadinstitute.org>'

# Default args: window size 3, guide size 2, allow GU pairing
# GU pairing allows AA to match GG in 1st window
SEQS = OrderedDict()
SEQS["genome_1"] = "AACTA"
SEQS["genome_2"] = "AAACT"
SEQS["genome_3"] = "GGCTA"
SEQS["genome_4"] = "GGCTT"

# Specificity seq stops AA from being the best guide in the 1st window
SP_SEQS = OrderedDict()
SP_SEQS["genome_5"] = "AA---"


class TestDesign(object):
    """General class for testing design.py

    Defines helper functions for test cases and basic setUp and
    tearDown functions.
    """
    class TestDesignCase(unittest.TestCase):
        def setUp(self):
            # Disable logging
            logging.disable(logging.INFO)

            # Create a temporary input file
            self.input_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
            # Closes the file so that it can be reopened on Windows
            self.input_file.close()

            # Create a temporary output file
            self.output_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
            self.output_file.close()

            self.files = [self.input_file.name, self.output_file.name]

        def check_results(self, file, expected, header='target-sequences'):
            """Check the results of the test output

            Given a TSV file of test output and expected output, fails the test
            if the test output guide target sequences do not equal the expected
            guide target sequences

            Args:
                file: string, path name of the file
                expected: list of lists of strings, all the expected guide
                    target sequences in each line of the output
                header: the header of the CSV that contains the guide target
                    sequences
            """
            col_loc = None
            with open(file) as f:
                for i, line in enumerate(f):
                    if i == 0:
                        headers = line.split('\t')
                        # Will raise an error if header is not in output
                        col_loc = headers.index(header)
                        continue
                    self.assertLess(i, len(expected) + 1)
                    guide_line = line.split('\t')[col_loc]
                    guides = guide_line.split(' ')
                    for guide in guides:
                        self.assertIn(guide, expected[i-1])
                    self.assertEqual(len(guides), len(expected[i-1]))
                self.assertEqual(i, len(expected))

        def baseArgv(self, search_type='sliding-window', input_type='fasta',
                     objective='minimize-guides', model=False, specific=None,
                     specificity_file=None, output_loc=None):
            """Get arguments for tests

            Produces the correct arguments for a test case given details of
            what the test case is testing. See design.py help for details
            on input

            Args:
                search_type: 'sliding-window' or 'complete-targets'
                input_type: 'fasta', 'auto-from-args', or 'auto-from-file'
                objective: 'minimize-guides' or 'maximize-activity'
                model: boolean, true to use Cas13a built in model, false
                    to use simple binary prediction
                specific: None, 'fasta', or 'taxa'; what sort of input
                    to be specific against
                output_loc: path to the output file/directory; set to
                    self.output_file.name if None

            Returns:
                List of strings that are the arguments of the test
            """
            input_file = self.input_file.name
            if output_loc is None:
                output_loc = self.output_file.name
            argv = ['design.py', search_type, input_type]

            if input_type == 'fasta':
                argv.extend([input_file, '-o', output_loc])
            elif input_type == 'auto-from-args':
                argv.extend(['64320', 'None', output_loc])
            elif input_type == 'auto-from-file':
                argv.extend([input_file, output_loc])

            if input_type in ['auto-from-args', 'auto-from-file']:
                argv.extend(['--sample-seqs', '1', '--mafft-path', 'fake_path'])

            if search_type == 'sliding-window':
                argv.extend(['-w', '3'])
            if search_type == 'complete-targets':
                argv.extend(['--best-n-targets', '2', '-pp', '.75', '-pl', '1',
                             '--max-primers-at-site', '2'])

            if objective == 'minimize-guides':
                argv.extend(['-gm', '0', '-gp', '.75'])
            elif objective =='maximize-activity':
                argv.extend(['--maximization-algorithm', 'greedy'])

            # ID-M (mismatches to be considered identical) must be set to 0 since otherwise
            # having 1 base in common with a 2 base guide counts as a match
            if specific == 'fasta':
                argv.extend(['--specific-against-fastas', specificity_file, '--id-m', '0'])
            elif specific == 'taxa':
                argv.extend(['--specific-against-taxa', specificity_file, '--id-m', '0'])

            if model:
                argv.append('--predict-cas13a-activity-model')
            elif objective =='maximize-activity':
                argv.extend(['--use-simple-binary-activity-prediction', '-gm', '0'])

            argv.extend(['--obj', objective, '--seed', '0', '-gl', '2'])

            return argv

        def tearDown(self):
            for file in self.files:
                if os.path.isfile(file):
                    os.unlink(file)
            # Re-enable logging
            logging.disable(logging.NOTSET)


class TestDesignFasta(TestDesign.TestDesignCase):
    """Test design.py given an input FASTA
    """

    def setUp(self):
        super().setUp()

        # Write to temporary input fasta
        seq_io.write_fasta(SEQS, self.input_file.name)

    def test_min_guides(self):
        argv = super().baseArgv()
        args = design.argv_to_args(argv)
        design.run(args)
        # Base args set the percentage of sequences to match at 75%
        expected = [["AA"], ["CT"], ["CT"]]
        self.check_results(self.output_file.name, expected)

    def test_max_activity(self):
        argv = super().baseArgv(objective='maximize-activity')
        args = design.argv_to_args(argv)
        design.run(args)
        # Doesn't use model, just greedy binary prediction with 0 mismatches
        # (so same outputs as min-guides)
        expected = [["AA"], ["CT"], ["CT"]]
        self.check_results(self.output_file.name, expected)

    def test_complete_targets(self):
        argv = super().baseArgv(search_type='complete-targets')
        args = design.argv_to_args(argv)
        design.run(args)
        # Since sequences are short and need 1 base for primer on each side,
        # only finds 1 target in middle
        expected = [["CT"]]
        self.check_results(self.output_file.name, expected,
                           header='guide-target-sequences')

    def test_specificity_fastas(self):
        # Create a temporary fasta file for specificity
        self.sp_fasta = tempfile.NamedTemporaryFile(mode='w', delete=False)
        # Closes the file so that it can be reopened on Windows
        self.sp_fasta.close()

        seq_io.write_fasta(SP_SEQS, self.sp_fasta.name)

        self.files.append(self.sp_fasta.name)

        argv = super().baseArgv(specific='fasta',
            specificity_file=self.sp_fasta.name)
        args = design.argv_to_args(argv)
        design.run(args)
        # AA isn't allowed in 1st window by specificity fasta,
        # so 1st window changes
        expected = [["AC", "GG"], ["CT"], ["CT"]]
        self.check_results(self.output_file.name, expected)


class TestDesignAutos(TestDesign.TestDesignCase):
    """Test design.py given arguments to automatically download FASTAs

    Does not run the entire design.py; prematurely stops by giving a fake path
    to MAFFT. All are expected to return a FileNotFoundError
    """
    def setUp(self):
        super().setUp()

        # Write to temporary input file
        with open(self.input_file.name, 'w') as f:
            f.write("Zika virus\t64320\tNone\tNC_035889\n")

        # Create a temporary output directory
        self.output_dir = tempfile.TemporaryDirectory()

    def test_auto_from_file(self):
        argv = super().baseArgv(input_type='auto-from-file',
                        output_loc=self.output_dir.name)
        args = design.argv_to_args(argv)
        try:
            design.run(args)
        except FileNotFoundError:
            pass

    def test_auto_from_args(self):
        argv = super().baseArgv(input_type='auto-from-args')
        args = design.argv_to_args(argv)
        try:
            design.run(args)
        except FileNotFoundError:
            pass

    def test_specificity_taxa(self):
        argv = super().baseArgv(input_type='auto-from-args',
                        specific='taxa', specificity_file='')
        args = design.argv_to_args(argv)
        try:
            design.run(args)
        except FileNotFoundError:
            pass

    def tearDown(self):
        super().tearDown()
        self.output_dir.cleanup()


class TestDesignFull(TestDesign.TestDesignCase):
    """Test design.py fully through
    """
    def setUp(self):
        super().setUp()

        # Write to temporary input file
        with open(self.input_file.name, 'w') as f:
            f.write("Zika virus\t64320\tNone\tNC_035889\n")

        # Create a temporary specificity file
        self.sp_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        self.sp_file.write("123\tNone\n")
        # Closes the file so that it can be reopened on Windows
        self.sp_file.close()

        # 'auto-from-args' gives different outputs for every cluster
        # Our test only produces 1 cluster, so store the name of that file
        self.real_output_file = self.output_file.name + '.0'
        self.files.extend([self.sp_file.name, self.real_output_file])

        # We cannot access MAFFT, so override this function; store original so
        # it can be fixed for future tests
        self.set_mafft_exec = align.set_mafft_exec
        align.set_mafft_exec = lambda mafft_path: None

        # Curating requires MAFFT, so override this function; store original so
        # it can be fixed for future tests
        self.curate_against_ref = align.curate_against_ref

        def small_curate(seqs, ref_accs, asm=None, remove_ref_accs=[]):
            return {seq: seqs[seq] for seq in seqs \
                if seq.split('.')[0] not in remove_ref_accs}

        align.curate_against_ref = small_curate

        # Aligning requires MAFFT, so override this function and output simple
        # test sequences; store original so it can be fixed for future tests
        self.align = align.align
        align.align = lambda seqs, am=None: SEQS

        # We don't want to fetch sequences for the specificity file since we're
        # doing a simple test case, so override this function; store original
        # so it can be fixed for future tests
        self.fetch_sequences_for_taxonomy = prepare_alignment.fetch_sequences_for_taxonomy

        def small_fetch(taxid, segment):
            # 123 is the taxonomic ID used in our specificity file
            if taxid == 123:
                return SP_SEQS
            # If it's not the specificity taxonomic ID, test fetching the real sequences,
            # although they won't be used
            else:
                return self.fetch_sequences_for_taxonomy(taxid, segment)

        prepare_alignment.fetch_sequences_for_taxonomy = small_fetch

    def test_specificity_taxa(self):
        argv = super().baseArgv(input_type='auto-from-args', specific='taxa',
            specificity_file=self.sp_file.name)
        args = design.argv_to_args(argv)
        design.run(args)
        # Same output as test_specificity_fasta, as sequences are the same
        expected = [["AC", "GG"], ["CT"], ["CT"]]
        self.check_results(self.real_output_file, expected)

    def tearDown(self):
        # Fix all overridden functions
        align.set_mafft_exec = self.set_mafft_exec
        align.curate_against_ref = self.curate_against_ref
        align.align = self.align
        prepare_alignment.fetch_sequences_for_taxonomy = self.fetch_sequences_for_taxonomy
        super().tearDown()



#!/usr/bin/env python3
"""Analyze coverage obtained by designs."""

import argparse
import logging

from adapt import coverage_analysis
from adapt.prepare import ncbi_neighbors
from adapt.utils import log
from adapt.utils import predict_activity
from adapt.utils import seq_io

__author__ = 'Hayden Metsky <hayden@mit.edu>'

logger = logging.getLogger(__name__)


def read_designs(fn):
    """Read a collection of targets from a file.

    Each target represents a design.

    Args:
        fn: path to a TSV file giving targets

    Returns:
        dict {design_id: design} where design_id is the row number
        of the target in fn (starting at 1 for the first target) and
        design is a coverage_analysis.Design object
    """
    rows = []
    with open(fn) as f:
        col_names = {}
        for i, line in enumerate(f):
            line = line.rstrip()
            ls = line.split('\t')
            if i == 0:
                # Parse header
                for j in range(len(ls)):
                    col_names[j] = ls[j]
            else:
                # Read each column as a variable
                cols = {}
                for j in range(len(ls)):
                    cols[col_names[j]] = ls[j]
                rows += [cols]

    designs = {}
    for i, row in enumerate(rows):
        cols = row
        if 'target-sequences' in cols:
            # This design only contains guides
            design = coverage_analysis.Design(
                    cols['target-sequences'].split(' '))
        else:
            # This design contains primers and guides
            design = coverage_analysis.Design(
                    cols['guide-target-sequences'].split(' '),
                    (cols['left-primer-target-sequences'].split(' '),
                     cols['right-primer-target-sequences'].split(' ')))
        designs[i + 1] = design
    return designs


def read_accessions(fn):
    """Read a list of accessions from a file.

    Args:
        fn: path to file where each line gives an accession

    Returns:
        collection of accessions
    """
    accs = []
    with open(fn) as f:
        for line in f:
            line = line.rstrip()
            accs += [line]
    return accs


def write_frac_bound(designs, frac_bound, out_fn):
    """Write table giving the fraction of sequences bound by a target.

    Args:
        designs: dict {design_id: design} where design_id is an
            identifier for design and design is a coverage_analysis.
            Design object
        frac_bound: dict {design_id: frac_bound} where design_id is
            an identifier for a design and frac_bound is the fraction
            of target sequences that are bound by the design
        out_fn: path to TSV file at which to write table
    """
    header = ['design_id',
              'guide-target-sequences',
              'left-primer-target-sequences',
              'right-primer-target-sequences',
              'frac-bound']
    with open(out_fn, 'w') as fw:
        fw.write('\t'.join(header) + '\n')
        for design_id in sorted(list(designs.keys())):
            guides = designs[design_id].guides
            guides = ' '.join(sorted(guides))
            if designs[design_id].primers is not None:
                left_primers, right_primers = designs[design_id].primers
                left_primers = ' '.join(sorted(left_primers))
                right_primers = ' '.join(sorted(right_primers))
            else:
                left_primers, right_primers = 'n/a', 'n/a'
            row = [design_id, guides, left_primers, right_primers,
                   frac_bound[design_id]]
            fw.write('\t'.join([str(x) for x in row]) + '\n')


def write_mean_activity_of_guides(designs, mean_activities, out_fn):
    """Write table giving the mean activity of guide sets.

    Args:
        designs: dict {design_id: design} where design_id is an
            identifier for design and design is a coverage_analysis.
            Design object
        mean_activities: dict {design_id: activity} where design_id is
            an identifier for a design and activity is the mean activity,
            across the target sequences, of its guide set
        out_fn: path to TSV file at which to write table
    """
    header = ['design_id',
              'guide-target-sequences',
              'mean-activity']
    with open(out_fn, 'w') as fw:
        fw.write('\t'.join(header) + '\n')
        for design_id in sorted(list(designs.keys())):
            guides = designs[design_id].guides
            guides = ' '.join(sorted(guides))
            row = [design_id, guides, mean_activities[design_id]]
            fw.write('\t'.join([str(x) for x in row]) + '\n')


def main(args):
    # Allow G-U base pairing, unless it is explicitly disallowed
    allow_gu_pairs = not args.do_not_allow_gu_pairing

    # Read the designs
    designs = read_designs(args.designs_fn)

    # If accessions were given, fetch them as use them as input
    if args.use_accessions:
        accessions = read_accessions(args.seqs_fn)
        seqs_tempfile = ncbi_neighbors.fetch_fastas(accessions)
        seqs_fn = seqs_tempfile.name
    else:
        seqs_tempfile = None
        seqs_fn = args.seqs_fn

    # Read the input sequences to compute coverage against; use
    # skip_gaps=True so that, if an alignment is input, this
    # is read as unaligned sequences
    seqs = seq_io.read_fasta(seqs_fn, skip_gaps=True)

    if (args.guide_mismatches is not None) and args.predict_activity_model_path:
        raise Exception(("Cannot set both --guide-mismatches and "
            "--predict-activity-model-path. Choose --guide-mismatches "
            "for a model based on mismatches, and --predict-activity-model-"
            "path to make determinations based on whether predicted "
            "activity is high."))
    elif args.guide_mismatches is not None:
        analyzer = coverage_analysis.CoverageAnalyzerWithMismatchModel(
                seqs, designs, args.guide_mismatches, args.primer_mismatches,
                allow_gu_pairs, fully_sensitive=args.fully_sensitive)
    elif args.predict_activity_model_path:
        cla_path, reg_path = args.predict_activity_model_path
        if args.predict_activity_thres:
            # Use specified thresholds on classification and regression
            cla_thres, reg_thres = args.predict_activity_thres
        else:
            # Use default thresholds specified with the model
            cla_thres, reg_thres = None, None
        predictor = predict_activity.Predictor(cla_path, reg_path,
                classification_threshold=cla_thres,
                regression_threshold=reg_thres)
        highly_active = args.predict_activity_require_highly_active
        analyzer = coverage_analysis.CoverageAnalyzerWithPredictedActivity(
                seqs, designs, predictor, args.primer_mismatches,
                highly_active=highly_active,
                fully_sensitive=args.fully_sensitive)
    else:
        raise Exception(("One of --guide-mismatches or "
            "--predict-activity-model-path must be set"))

    # Perform analyses
    performed_analysis = False
    if args.write_frac_bound:
        frac_bound = analyzer.frac_of_seqs_bound()
        write_frac_bound(designs, frac_bound, args.write_frac_bound)
        performed_analysis = True
    if args.write_mean_activity_of_guides:
        if (not args.predict_activity_model_path or
                args.predict_activity_require_highly_active):
            raise Exception(("To use --write-mean-activity-of-guides, "
                    "a predictive model must be set and "
                    "--predict-activity-require-highly-active must *not* "
                    "be set"))
        mean_activity = analyzer.mean_activity_of_guides()
        write_mean_activity_of_guides(designs, mean_activity,
                args.write_mean_activity_of_guides)
        performed_analysis = True

    if not performed_analysis:
        logger.warning(("No analysis was requested"))

    # Close tempfiles
    if seqs_tempfile is not None:
        seqs_tempfile.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # Required inputs
    parser.add_argument('designs_fn',
        help=("Path to output of running design.py; a TSV file where "
              "each row contains a design (target)"))
    parser.add_argument('seqs_fn',
        help=("Path to FASTA file giving sequences against which to "
              "compute coverage. (See --use-accessions to pass accessions "
              "as input rather than a FASTA file.)"))

    # Analyses to output
    parser.add_argument('--write-frac-bound',
        help=("If set, write a table in which each row represents an "
              "input design and gives the fraction of all sequences that "
              "are covered by the design. The 'design_id' column gives "
              "the row number of the design in the designs input (1 for "
              "the first design). The provided argument is a path to "
              "a TSV file at which to the write the table."))
    parser.add_argument('--write-mean-activity-of-guides',
        help=("If set, write a table in which each row represents an "
              "input design and gives the mean activity across the target "
              "sequences of the guide set. The 'design_id' column gives "
              "the row number of the design in the designs input (1 for "
              "the first design). The provided argument is a path to "
              "a TSV file at which to write the table. If set, a predictive "
              "model must be set without "
              "--predict-activity-require-highly-active"))

    # Parameter determining whether a primer binds to target
    parser.add_argument('-pm', '--primer-mismatches',
        type=int, default=0,
        help=("Allow for this number of mismatches when determining "
              "whether a primer covers a sequence (ignore this if "
              "the targets only consist of guides)"))

    # Parameters determining whether a guide binds to target based on
    # mismatch model
    parser.add_argument('-gm', '--guide-mismatches',
        type=int,
        help=("Allow for this number of mismatches when "
              "determining whether a guide covers a sequence; either this "
              "or --predict-activity-model-path should be set"))
    parser.add_argument('--do-not-allow-gu-pairing',
        action='store_true',
        help=("When determining whether a guide binds to a region of "
              "target sequence, do not count G-U (wobble) base pairs as "
              "matching. Default is to tolerate G-U pairing: namely, "
              "A in an output guide sequence matches G in the "
              "target and C in an output guide sequence matches T "
              "in the target (since the synthesized guide is the reverse "
              "complement of the output guide sequence)"))

    # Parameters determining whether a guide binds to target based on
    # trained model
    parser.add_argument('--predict-activity-model-path',
        nargs=2,
        help=("Paths to directories containing serialized models in "
              "TensorFlow's SavedModel format for predicting guide-target "
              "activity. There are two arguments: (1) classification "
              "model to determine which guides are active; (2) regression "
              "model, which is used to determine which guides (among "
              "active ones) are highly active. The models/ directory "
              "contains example models. Either this or --guide-mismatches "
              "should be set."))
    parser.add_argument('--predict-activity-thres',
        type=float,
        nargs=2,
        help=("Thresholds to use for decisions on output of predictive "
            "models. There are two arguments: (1) classification threshold "
            "for deciding which guide-target pairs are active (in [0,1], "
            "where higher values have higher precision but less recall); "
            "(2) regression threshold for deciding which guide-target pairs "
            "are highly active (>= 0, where higher values limit the number "
            "determined to be highly active). If not set but --predict-"
            "activity-model-path is set, then this uses default thresholds "
            "stored with the models. To 'bind to' or 'cover' a target, "
            "the guide-target pair must be active or, if "
            "--predict-activity-require-highly-active is set, highly active."))
    parser.add_argument('--predict-activity-require-highly-active',
        action='store_true',
        help=("When determining whether a guide-target pair binds using an "
              "activity model, require that the pair be predicted to be "
              "highly active (not just active)"))

    # Miscellaneous
    parser.add_argument('--use-accessions',
        action='store_true',
        help=("When set, the input file of sequences gives accessions rather "
              "than being a FASTA of sequences -- each line in the file gives "
              "an accession. This fetches the sequences of those accessions "
              "uses them as input."))
    parser.add_argument('--fully-sensitive',
        action='store_true',
        help=("When set, use a naive, slow sliding approach to find binding "
              "for primers and guides; otherwise, this uses an index to "
              "more quickly identify binding sites"))

    # Log levels
    parser.add_argument("--debug",
        dest="log_level",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
        help=("Debug output"))
    parser.add_argument("--verbose",
        dest="log_level",
        action="store_const",
        const=logging.INFO,
        help=("Verbose output"))

    args = parser.parse_args()
    log.configure_logging(args.log_level)
    main(args)



"""Tests for ncbi_neighbors module.
"""

import unittest

from adapt.prepare import ncbi_neighbors as nn

__author__ = 'Hayden Metsky <hayden@mit.edu>'


class TestURLConstruction(unittest.TestCase):
    """Tests constructing URLs.
    """

    def test_ncbi_neighbors_url(self):
        url = nn.ncbi_neighbors_url(123)
        expected_url = ('https://www.ncbi.nlm.nih.gov/genomes/GenomesGroup'
            '.cgi?taxid=123&cmd=download2')
        self.assertEqual(url, expected_url)

    def test_ncbi_reference_url(self):
        url = nn.ncbi_reference_url(123)
        expected_url = ('https://www.ncbi.nlm.nih.gov/genomes/GenomesGroup'
            '.cgi?taxid=123&cmd=download1')
        self.assertEqual(url, expected_url)

    def test_ncbi_fasta_download_url(self):
        url = nn.ncbi_fasta_download_url(['A123', 'A456', 'B789'])
        expected_url = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch'
            '.fcgi?id=A123,A456,B789&db=nuccore&rettype=fasta&retmode=text')
        self.assertEqual(url, expected_url)

    def test_ncbi_xml_download_url(self):
        url = nn.ncbi_xml_download_url(['A123', 'A456', 'B789'])
        expected_url = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch'
            '.fcgi?id=A123,A456,B789&db=nuccore&retmode=xml')
        self.assertEqual(url, expected_url)

    def test_read_taxid(self):
        url = ('https://www.ncbi.nlm.nih.gov/genomes/GenomesGroup'
            '.cgi?taxid=123&cmd=download2')
        taxid = nn.read_taxid_from_ncbi_url(url)
        expected_taxid = 123
        self.assertEqual(taxid, expected_taxid)


class TestConstructNeighbors(unittest.TestCase):
    """Tests the construct_neighbors() function.

    The function construct_neighbors() calls fetch_neighbors_table(),
    which makes a request to NCBI. To avoid the request, this overrides
    fetch_neighbors_table() to return a known neighbors table.
    """

    def setUp(self):
        self.expected_table_contents = \
            ("## Comment line 1\n"
             "## Comment line 2\n"
             "## Columns:\tRepresentative\tNeighbor\tHost\tSelected lineage\tTaxonomy name\tSegment name\n"
             "NC_0123\tKY456\tvertebrate,human\tFamilyA,GenusA,SpeciesA\tSpeciesA\tsegment \n"
             "NC_0456\tAB123\tinvertebrate\tFamilyB,GenusB,SpeciesB\tSpeciesB\tsegment 1\n"
             "NC_0456\tAB456\tinvertebrate\tFamilyB,GenusB,SpeciesB\tSpeciesB\tsegment 2\n")
        self.expected_neighbors = [
            nn.Neighbor('KY456', 'NC_0123', ['vertebrate', 'human'],
                ('FamilyA', 'GenusA', 'SpeciesA'), 'SpeciesA', ''),
            nn.Neighbor('AB123', 'NC_0456', ['invertebrate'],
                ('FamilyB', 'GenusB', 'SpeciesB'), 'SpeciesB', '1'),
            nn.Neighbor('AB456', 'NC_0456', ['invertebrate'],
                ('FamilyB', 'GenusB', 'SpeciesB'), 'SpeciesB', '2')
        ]

        # Override nn.fetch_neighbors_table() to return expected_table_lines,
        # but keep the real function
        self.fetch_neighbors_table_real = nn.fetch_neighbors_table
        nn.fetch_neighbors_table = lambda taxid: self.expected_table_contents.split('\n')

    def test_construct_neighbors(self):
        neighbors = nn.construct_neighbors(123)
        self.assertEqual(neighbors, self.expected_neighbors)

    def tearDown(self):
        # Reset nn.fetch_neighbors_table()
        nn.fetch_neighbors_table = self.fetch_neighbors_table_real
